import pickle
import random
import time

from google.appengine.ext import db

from connection import Connection, ConnectionError
from connection_manager import ConnectionManager, NotConnectedError, SessionAttrError

class GaeConnection(Connection):
    """A connection stored in Google Datastore."""

    def __init__(self, *args, **kwargs):
        Connection.__init__(self, *args, **kwargs)
        self.model = None

class GaeConnectionLastActive(db.Model):
    value = db.FloatProperty(required=True)

class GaeConnectionConnected(db.Model):
    value = db.BooleanProperty(required=True)

class GaeConnectionLastPolled(db.Model):
    value = db.FloatProperty(required=True)

class GaeConnectionAuthentication(db.Model):
    authenticated = db.BooleanProperty(required=True)
    flex_user = db.StringProperty(required=False)

class GaeConnectionSession(db.Model):
    value = db.BlobProperty(required=True)

class GaeConnectionModel(db.Model):
    """Connection data that is stored in a Google Datastore."""

    # These values never change
    channel_name = db.StringProperty(required=True)
    timeout = db.IntegerProperty(required=True)

    # Every changeable property is it's own model,
    # So we can update each property
    # without affecting any other.

    # Referenced properties.
    last_active = db.ReferenceProperty(reference_class=GaeConnectionLastActive, required=True)
    connected = db.ReferenceProperty(reference_class=GaeConnectionConnected, required=True)
    last_polled = db.ReferenceProperty(reference_class=GaeConnectionLastPolled, required=True)
    authentication = db.ReferenceProperty(reference_class=GaeConnectionAuthentication, required=False)
    session = db.ReferenceProperty(reference_class=GaeConnectionSession, required=False)

class GaeChannelModel(db.Model):
    """Channel data that is stored in a Google Datastore."""

    # Shard counter into multiple entities to avoid conflicts.
    NUM_SHARDS = 20

    name = db.StringProperty(required=True)
    count = db.IntegerProperty(required=True)

class GaeConnectionManager(ConnectionManager):
    """Manages connections stored by Google DataStore.

    attributes
    =============
     * touch_time - int, minimum time in milliseconds between writes to 'last_active' field.
    """

    def __init__(self, connection_class=GaeConnection, connection_params=None, touch_time=10000):
        ConnectionManager.__init__(self, connection_class=connection_class,
            connection_params=connection_params)

        # Reduces number of writes to 'last_active' field.
        self.touch_time = touch_time

    def reset(self):
        query = GaeConnectionModel.all(keys_only=True)
        for result in query:
            db.delete(result)

        query = GaeConnectionLastActive.all(keys_only=True)
        for result in query:
            db.delete(result)

        query = GaeConnectionConnected.all(keys_only=True)
        for result in query:
            db.delete(result)

        query = GaeConnectionLastPolled.all(keys_only=True)
        for result in query:
            db.delete(result)

        query = GaeConnectionAuthentication.all(keys_only=True)
        for result in query:
            db.delete(result)

        query = GaeConnectionSession.all(keys_only=True)
        for result in query:
            db.delete(result)

        query = GaeChannelModel.all(keys_only=True)
        for result in query:
            db.delete(result)

    def _getChannelShardName(self, channel_name):
        index = random.randint(0, GaeChannelModel.NUM_SHARDS - 1)
        return ":".join((channel_name, str(index)))

    def _incrementChannelCount(self, channel_name):
        shard_name = self._getChannelShardName(channel_name)
        counter = GaeChannelModel.get_by_key_name(shard_name)
        if counter is None:
            counter = GaeChannelModel(key_name=shard_name,
                name=channel_name, count=0)
        counter.count += 1
        counter.put()

    def _decrementChannelCount(self, channel_name):
        shard_name = self._getChannelShardName(channel_name)
        counter = GaeChannelModel.get_by_key_name(shard_name)
        if counter is None:
            counter = GaeChannelModel(key_name=shard_name,
                name=channel_name, count=0)
        counter.count -= 1
        counter.put()

    def getConnectionCount(self, channel_name):
        query = GaeChannelModel.all()
        query.filter('name = ', channel_name)
        total = 0
        for result in query:
            total += result.count
        return total
 
    def loadConnection(self, connection_id):
        connection_model = GaeConnectionModel.get_by_key_name(connection_id)
        
        if connection_model is None:
            raise NotConnectedError("Connection '%s' is not connected." % connection_id)

        connection = self.connection_class(self, connection_model.channel_name,
            connection_id, timeout=connection_model.timeout)
        connection.model = connection_model
        return connection

    def initConnection(self, connection, channel):
        last_active = GaeConnectionLastActive(key_name=connection.id,
            value=(time.time() * 1000))
        last_active.put()

        connected = GaeConnectionConnected(key_name=connection.id, value=True)
        connected.put()

        last_polled = GaeConnectionLastPolled(key_name=connection.id, value=0.0)
        last_polled.put()

        params = {
            'key_name': connection.id,
            'channel_name': connection.channel_name,
            'timeout': connection.timeout,
            'connected': connected,
            'last_active': last_active,
            'last_polled': last_polled
        }

        connection_model = GaeConnectionModel(**params)
        connection_model.put()

        db.run_in_transaction(self._incrementChannelCount,
            connection.channel_name)

        connection.model = connection_model

    def iterConnectionIds(self):
        query = GaeConnectionModel.all(keys_only=True)
        for key in query:
            yield key.name()

    # --- proxies for connection properties --- #

    def getConnected(self, connection):
        if connection.model is None:
            return False
        else:
            return connection.model.connected.value

    def getLastActive(self, connection):
        return connection.model.last_active.value

    def getLastPolled(self, connection):
        return connection.model.last_polled.value

    def getAuthenticated(self, connection):
        if connection.model.authentication is None:
            return False
       
        return connection.model.authentication.authenticated

    def getFlexUser(self, connection):
        if connection.model.authentication is None:
            return None

        return connection.model.authentication.flex_user

    def getNotifyFunc(self, connection):
        return None

    # --- proxies for connection methods --- #

    def deleteConnection(self, connection):
        # Delete referenced properties 1st
        db.delete(GaeConnectionModel.last_active.get_value_for_datastore(connection.model))
        db.delete(GaeConnectionModel.connected.get_value_for_datastore(connection.model))
        db.delete(GaeConnectionModel.last_polled.get_value_for_datastore(connection.model))

        # Optional referenced properties
        authentication_key = GaeConnectionModel.authentication.get_value_for_datastore(connection.model)
        if authentication_key is not None:
            db.delete(authentication_key)

        session_key = GaeConnectionModel.session.get_value_for_datastore(connection.model)
        if session_key is not None:
            db.delete(session_key)

        # Delete connection
        connection.model.delete()
        connection.model = None
        ConnectionManager.deleteConnection(self, connection)

        db.run_in_transaction(self._decrementChannelCount, connection.channel_name)

    def connectConnection(self, connection):
        if connection.model is not None:
            connected = GaeConnectionConnected(key_name=connection.id, value=True)
            connected.put()
            connection.model.connected = connected

    def disconnectConnection(self, connection):
        if connection.model is not None:
            connected = GaeConnectionConnected(key_name=connection.id, value=False)
            connected.put()
            connection.model.connected = connected

    def touchConnection(self, connection):
        if connection.model is not None:
            now = time.time() * 1000
            diff = now - connection.model.last_active.value
            connection.model.last_active.value = now
            if diff > self.touch_time:
                # last_active value is only written periodically
                # to save time writing to data_store.
                connection.model.last_active.put()

    def touchPolled(self, connection):
        self.softTouchPolled(connection);
        if connection.model is not None:
            connection.model.last_polled.put()

    def softTouchPolled(self, connection):
        if connection.model is not None:
            connection.model.last_polled.value = time.time() * 1000

    def authenticateConnection(self, connection, user):
        if connection.model is not None:
            if connection.model.authentication is None:
                authentication = GaeConnectionAuthentication(
                    key_name=connection.id, authenticated=True, flex_user=user)
                authentication.put()
                connection.model.authentication = authentication
                connection.model.put()
            else:
                connection.model.authentication.authenticated = True
                connection.model.authentication.flex_user = user
                connection.model.authentication.authenticated.put()

    def unAuthenticateConnection(self, connection):
        if connection.model is not None:
            if connection.model.authentication is not None:
                connection.model.authentication.authenticated = False
                connection.model.authentication.flex_user = None
                connection.model.authentication.put()

    def initSession(self, connection):
        if connection.model is not None:
            if connection.model.session is None or \
                connection.model.session.value is None:
                connection._session = {}
            else:
                connection._session = pickle.loads(connection.model.session.value)

    def saveSession(self, connection):
        if connection.model is not None:
            value = pickle.dumps(connection._session)
            if connection.model.session is None:
                session = GaeConnectionSession(key_name=connection.id, value=value)
                session.put()
                connection.model.session = session
                connection.model.put()
            else:
                connection.model.session.value = value
                connection.model.session.put()

    def getConnectionSessionAttr(self, connection, name):
        self.initSession(connection)
        try:
            return connection._session[name]
        except KeyError:
            raise SessionAttrError("Attribute '%s' not found." % name)

    def setConnectionSessionAttr(self, connection, name, val):
        self.initSession(connection)
        connection._session[name] = val
        self.saveSession(connection)

    def delConnectionSessionAttr(self, connection, name):
        self.initSession(connection)
        try:
            del connection._session[name]
            self.saveSession(connection)
        except KeyError:
            pass



import optparse
from timeit import Timer

try:
    import pyamf
    from pyamf.util import BufferedByteStream
    use_pyamf = True
except ImportError:
    use_pyamf = False

from amfast.context import DecoderContext, EncoderContext
import amfast.encode as encode
import amfast.decode as decode
import amfast.class_def as class_def

class SpeedTestCase():
    class TestObject(object):
        STRING_REF = 'string reference'
        
        def __init__(self):
            self.null = None
            self.test_list = ['test', 'tester']
            self.test_dict = {'test': 'ignore'}

    class TestSubObject(object):
        def __init__(self):
            self.number = None

    class TestStaticObject(object):
        class __amf__:
            static = ('name', 'score', 'rank')

        def __init__(self, name='', score='', rank=''):
            self.name = name
            self.score = score
            self.rank = rank

    def setUp(self):
        self.class_mapper = class_def.ClassDefMapper()

        self.class_mapper.mapClass(class_def.DynamicClassDef(self.TestObject,
            'test_complex.test', (), amf3=False))
        self.class_mapper.mapClass(class_def.DynamicClassDef(self.TestSubObject,
            'test_complex.sub', (), amf3=False))
        self.class_mapper.mapClass(class_def.ClassDef(self.TestStaticObject,
            'test_complex.static', ('name', 'score', 'rank'), amf3=False))

        if use_pyamf is True:
            pyamf.register_class(self.TestObject, 'test_complex.test')
            pyamf.register_class(self.TestSubObject, 'test_complex.sub')
            pyamf.register_class(self.TestStaticObject, 'test_complex.static')

    def tearDown(self):
        self.class_mapper.unmapClass(self.TestObject)
        self.class_mapper.unmapClass(self.TestSubObject)

        if use_pyamf is True:
            pyamf.unregister_class(self.TestObject)
            pyamf.unregister_class(self.TestSubObject)

    def buildSimple(self, max=5):
        test_objects = []
        
        for i in xrange(0, max):
            test_obj = {
                'number': 10,
                'float': 3.24,
                'string': 'foo number %i' % i
            }
            test_objects.append(test_obj)
        return test_objects

    def buildComplex(self, max=5):
        test_objects = []

        for i in xrange(0, max):
            test_obj = self.TestObject()
            test_obj.number = i
            test_obj.float = 3.14
            test_obj.unicode = u'spam'
            test_obj.str = 'a l' + 'o' * 500 + 'ng string'
            test_obj.str_ref = self.TestObject.STRING_REF
            test_obj.sub_obj = self.TestSubObject()
            test_obj.sub_obj.number = i
            test_obj.ref = test_obj.sub_obj
            test_objects.append(test_obj)

        return test_objects

    def buildBig(self, object_size=5, object_count=1000):
        test_objects = []
        for i in xrange(0, object_count):
            test_objects.append(self.buildSimple())

    def buildStatic(self, max=5):
        test_objects = []

        for i in xrange(0, max):
            test_objects.append(self.TestStaticObject('name %s' % i, 5.5555555, i))
        return test_objects

    def buildReference(self, object_size=1000):
        obj = {'foo': 'bar'}

        objs = []
        for i in xrange(0, object_size):
            objs.append(obj)
 
    def encode(self, obj, pyamf=False, amf3=False, use_proxies=False):
        if pyamf is True:
            return self.pyamfEncode(obj, amf3, use_proxies)
        else:
            return self.amfastEncode(obj, amf3, use_proxies)

    def amfastEncode(self, obj, amf3=False, use_proxies=False):        
        enc_context = EncoderContext(use_collections=use_proxies,
            use_proxies=use_proxies, class_def_mapper=self.class_mapper,
            amf3=amf3)

        return encode.encode(obj, enc_context)

    def decode(self, str, amf3=False, pyamf=False):
        if pyamf is True:
            return self.pyamfDecode(str, amf3)
        else:
            return self.amfastDecode(str, amf3)

    def amfastDecode(self, str, amf3=False):
        decoded = decode.decode(DecoderContext(str,
            class_def_mapper=self.class_mapper, amf3=amf3))

    def pyamfEncode(self, obj, amf3=False, use_proxies=False):
        if amf3 is True:
            context = pyamf.get_context(pyamf.AMF3)
        else:
            context = pyamf.get_context(pyamf.AMF0)

        stream = BufferedByteStream()

        if amf3 is True:
            pyamf_encoder = pyamf.get_encoder(pyamf.AMF3, stream=stream, context=context)
        else:
            pyamf_encoder = pyamf.get_encoder(pyamf.AMF0, stream=stream, context=context)

        pyamf_encoder.writeElement(obj)
        return pyamf_encoder.stream.getvalue()

    def pyamfDecode(self, str, amf3=False):
        if amf3 is True:
            context = pyamf.amf3.Context()
            decoded = pyamf.amf3.Decoder(str, context).readElement()
        else:
            context = pyamf.amf0.Context()
            decoded = pyamf.amf0.Decoder(str, context).readElement()

    def simple(self, decode=False, amf3=False, pyamf=False):
        self.doTest(self.buildSimple, 50000, decode=decode, amf3=amf3, pyamf=pyamf)

    def complex(self, decode=False, amf3=False, pyamf=False):
        self.doTest(self.buildComplex, 10000, decode=decode, amf3=amf3, pyamf=pyamf)

    def big(self, decode=False, amf3=False, pyamf=False):
        self.doTest(self.buildBig, 1000, decode=decode, amf3=amf3, pyamf=pyamf)

    def static(self, decode=False, amf3=False, pyamf=False):
        self.doTest(self.buildStatic, 50000, decode=decode, amf3=amf3, pyamf=pyamf)

    def reference(self, decode=False, amf3=False, pyamf=False):
        self.doTest(self.buildReference, 25000, decode=decode, amf3=amf3, pyamf=pyamf)

    def doTest(self, obj_func, obj_count=10, decode=False, amf3=False, pyamf=False):
        for i in xrange(obj_count):
            obj = obj_func()
            encoded = self.encode(obj, pyamf=pyamf, amf3=amf3, use_proxies=False)
            if decode is True:
                decoded = self.decode(encoded, pyamf=pyamf, amf3=amf3)

    def benchmarkTest(self, function_name, label, repeat=3, print_result=True):
        self.setUp()

        t = Timer("%s()" % function_name, "from __main__ import %s" % function_name)
        e_time = t.timeit(repeat)

        if print_result is True:
            print "----"
            print "Running %s %ix:" % (label, repeat)
            print "%s seconds" % e_time
            print "----"

        self.tearDown()
        return e_time

    def runBenchmarks(self, tests, results, pyamf=False, repeat=3, print_result=True):
        for test in tests:
            def _test():
                test['test'](*test['args'], **test['kwargs'])

            global current_test
            current_test = _test

            if pyamf is True:
                label = 'PyAmf: ' + test['label']
            else:
                label = 'AmFast: ' + test['label']
            e_time = tester.benchmarkTest('current_test', label, repeat, print_result)

            if pyamf is True:
                result = results[test['label']]
                result['PyAmf'] = e_time
                result['diff'] = e_time / result['AmFast']
            else:
                results[test['label']] = {'AmFast': e_time}


if __name__ == "__main__":
    usage = """usage: %s [options]""" % __file__
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("-r", "--repeat", dest="repeat", default=1, help="Number of times to repeat each test.")
    parser.add_option("-p", action="store_true", dest="print_result", help="Print live output")
    parser.add_option("-a", action="store_true", dest="pyamf", help="Test PyAmf")
    (options, args) = parser.parse_args()

    tester = SpeedTestCase()
    tests = [
        {'test': tester.simple, 'label': 'AMF0 Encode Simple',
            'args': [], 'kwargs': {'decode': False, 'amf3': False, 'pyamf': False}},
        {'test': tester.simple, 'label': 'AMF0 Encode/Decode Simple',
            'args': [], 'kwargs': {'decode': True, 'amf3': False, 'pyamf': False}},

        {'test': tester.complex, 'label': 'AMF0 Encode Complex',
            'args': [], 'kwargs': {'decode': False, 'amf3': False, 'pyamf': False}},
        {'test': tester.complex, 'label': 'AMF0 Encode/Decode Complex',
            'args': [], 'kwargs': {'decode': True, 'amf3': False, 'pyamf': False}},


        {'test': tester.big, 'label': 'AMF0 Encode Big',
            'args': [], 'kwargs': {'decode': False, 'amf3': False, 'pyamf': False}},
        {'test': tester.big, 'label': 'AMF0 Encode/Decode Big',
            'args': [], 'kwargs': {'decode': True, 'amf3': False, 'pyamf': False}},

        {'test': tester.static, 'label': 'AMF0 Encode Static',
            'args': [], 'kwargs': {'decode': False, 'amf3': False, 'pyamf': False}},
        {'test': tester.static, 'label': 'AMF0 Encode/Decode Static',
            'args': [], 'kwargs': {'decode': True, 'amf3': False, 'pyamf': False}},

        {'test': tester.reference, 'label': 'AMF0 Encode Reference',
            'args': [], 'kwargs': {'decode': False, 'amf3': False, 'pyamf': False}},
        {'test': tester.reference, 'label': 'AMF0 Encode/Decode Reference',
            'args': [], 'kwargs': {'decode': True, 'amf3': False, 'pyamf': False}},
    ]

    # Add AMF3 tests
    amf3_tests = []
    for test in tests:
        amf3_test = {
            'test': test['test'],
            'label': test['label'].replace('AMF0', 'AMF3'),
            'args': [],
            'kwargs': {}
        }


        amf3_test['args'].extend(test['args'])
        amf3_test['kwargs'].update(test['kwargs'])
        amf3_test['kwargs']['amf3'] = True
        amf3_tests.append(amf3_test);
    tests.extend(amf3_tests)

    # Run AmFast tests
    results = {}
    tester.runBenchmarks(tests, results, pyamf=False, repeat=options.repeat, print_result=options.print_result)

    if options.pyamf:
        pyamf_tests = []
        for test in tests:
            pyamf_test = {
                'test': test['test'],
                'label': test['label'],
                'args': [],
                'kwargs': {}
            }

            pyamf_test['args'].extend(test['args'])
            pyamf_test['kwargs'].update(test['kwargs'])
            pyamf_test['kwargs']['pyamf'] = True
            pyamf_tests.append(pyamf_test);

        # Run PyAmf tests
        tester.runBenchmarks(pyamf_tests, results, pyamf=True, repeat=options.repeat, print_result=options.print_result)

    # Print results
    header = ('', 'AmFast')
    if options.pyamf:
        header.extend(['PyAmf', 'Diff'])
    print ','.join(header)
    for label, result in results.iteritems():
        row = [label, '%s' % result['AmFast']]
        if options.pyamf:
            row.extend(['%s' % result['PyAmf'], '%s' % result['diff']])
        print ','.join(row)



"""Equivalent to Flex mx.messaging.messages package."""
import uuid
import time
import cgi

import amfast
from amfast import class_def, remoting
from amfast.class_def.as_types import AsError

try:
    # Use decode module if available.
    # Users may be using PyAmf instead.
    from amfast.decode import decode
except ImportError:
    pass

class FlexMessageError(remoting.RemotingError):
    """Errors raised by this module."""
    pass

class FaultError(AsError):
    """Equivalent to mx.rpc.Fault."""

    def __init__(self, message='', exc=None, detail='', content=None):
        AsError.__init__(self, message, exc)
        
        self.faultCode = self.name
        self.faultString = self.message
        self.faultDetail = detail
        self.rootCause = exc
        self.content = content
class_def.assign_attrs(FaultError, 'mx.rpc.Fault',
    ('errorId', 'name', 'message', 'faultCode',
        'faultString', 'faultDetail', 'rootCause', 'content'), True)

class AbstractMessage(object):
    """Base class for all FlexMessages."""

    DESTINATION_CLIENT_ID_HEADER = 'DSDstClientId'
    ENDPOINT_HEADER = 'DSEndpoint'
    FLEX_CLIENT_ID_HEADER = 'DSId'
    PRIORITY_HEADER = 'DSPriority'
    REMOTE_CREDENTIALS_CHARSET_HEADER = 'DSRemoteCredentialsCharset'
    REMOTE_CREDENTIALS_HEADER = 'DSRemoteCredentials'
    REQUEST_TIMEOUT_HEADER = 'DSRequestTimeout'
    STATUS_CODE_HEADER = 'DSStatusCode'

    def __init__(self, body=None, clientId=None, destination=None,
        headers=None, timeToLive=None, timestamp=None, messageId=None):
        self.body = body
        self.clientId = clientId
        self.destination = destination
        self.timeToLive = timeToLive

        if headers is not None:
            self.headers = headers

        if timestamp is None:
            timestamp = time.time() * 1000
        self.timestamp = timestamp
   
        if messageId is None:
            messageId = self._getId()
        self.messageId = messageId

    def invoke(self, packet, msg):
        """Invoke all message headers."""
        if amfast.log_debug:
            amfast.logger.debug("\nInvoking FlexMessage:\n%s" % self)

    def fail(self, packet, msg, exc):
        """Return an error message."""
        fault = FaultError(exc=exc)
        response = ErrorMessage(exc=fault)
        self._matchAcknowledge(packet, msg, response)
        return response

    def convertFail(self, exc):
        """Convert this message to an error."""
        fault = FaultError(exc=exc)
        headers = getattr(self, 'headers', None)
        return ErrorMessage(self, clientId=self.clientId, destination=self.destination,
            headers=headers, timeToLive=self.timeToLive, timestamp=self.timestamp,
            messageId=self.messageId, correlationId=self.correlationId, exc=fault)

    def getAcknowledgeClass(self):
        """Returns the correct class for the response message."""
        return AcknowledgeMessage

    def acknowledge(self, packet, msg):
        """Return a successful result message."""
        class_ = self.getAcknowledgeClass()
        response = class_()
        self._matchAcknowledge(packet, msg, response)
        return response

    def _matchAcknowledge(self, packet, msg, response):
        """Syncs values between this message and it's response acknowledgement."""
        response.correlationId = self.messageId

        if self.clientId is None:
            self.clientId = packet.channel.channel_set.connection_manager.generateId()
        response.clientId = self.clientId

    def _getId(self):
        """Get a messageId or clientId."""
        return str(uuid.uuid4())

    def __str__(self):
        header_str = ''
        if hasattr(self, 'headers') and self.headers is not None:
            header_str = '\n  '.join(["<header name=\"%s\">%s</header>" % (key, val) for key, val in self.headers.iteritems()])
        
        attrs = {}
        for key, val in self.__dict__.iteritems():
            if key == 'body':
                continue
            if key == 'headers':
                continue
            attrs[key] = val
        
        attrs_str = '\n  '.join(["<attr name=\"%s\">%s</attr>" % (key, val) for key, val in attrs.iteritems()])

        str = """
<FlexMessage: %s>
 <headers>
  %s
 </headers>

 <body>
  %s
 </body>

 <attributes>
  %s
 </attributes>
</FlexMessage>
""" % (self.__class__.__name__, header_str, self.body, attrs_str)
        return str

class_def.assign_attrs(AbstractMessage, 'flex.messaging.messages.AbstractMessage',
    ('body', 'clientId', 'destination', 'headers',
        'messageId', 'timestamp', 'timeToLive'), True)

class AbstractSmallMsgDef(class_def.ExternClassDef):
    """Encodes and decodes messages using ISmallMessage.

    ISmallMessages use a more compact representation
    of mx.messaging.messages.
    """

    HAS_NEXT_FLAG = 0x80
    BODY_FLAG = 0x01
    CLIENT_ID_FLAG = 0x02
    DESTINATION_FLAG = 0x04
    HEADERS_FLAG = 0x08
    MESSAGE_ID_FLAG = 0x10
    TIMESTAMP_FLAG = 0x20
    TIME_TO_LIVE_FLAG = 0x40
    CLIENT_ID_BYTES_FLAG = 0x01
    MESSAGE_ID_BYTES_FLAG = 0x02

    ALPHA_CHAR_CODES = (48, 49, 50, 51, 52, 53, 54, 
        55, 56, 57, 65, 66, 67, 68, 69, 70)

    def _readUid(self, bytes):
        """Decode a 128bit byte array into a 36 char string representing an UID."""
        if bytes is None:
            return None

        if hasattr(bytes, 'bytes'):
            # amfast.class_def.as_types.ByteArray object
            byte_str = bytes.bytes
        else:
            # Other type
            byte_str = str(bytes)

        if len(byte_str) != 16:
           return None

        uid_chars = [None] * 36
        idx = 0
        for i, byte in enumerate(byte_str):
            if i == 4 or i == 6 or i == 8 or i == 10:
                # hyphen
                uid_chars[idx] = 45
                idx += 1

            char_code = ord(byte)
            uid_chars[idx] = self.ALPHA_CHAR_CODES[(char_code & 0xF0) >> 4]
            idx += 1
            uid_chars[idx] = self.ALPHA_CHAR_CODES[(char_code & 0x0F)]
            idx += 1
        return ''.join([chr(byte) for byte in uid_chars])

    def _readFlags(self, context):
        """Reads flags."""
        flags = []

        flag = self.HAS_NEXT_FLAG
        while (flag & self.HAS_NEXT_FLAG):
            flag = ord(context.read(1))
            flags.append(flag)

        return flags

    def readExternal(self, obj, context):
        flags = self._readFlags(context)

        for i, flag in enumerate(flags):
            if i == 0:
                if flag & self.BODY_FLAG:
                    obj.body = decode(context)
                else:
                    obj.body = None

                if flag & self.CLIENT_ID_FLAG:
                    obj.clientId = decode(context)
                else:
                    obj.clientId = None

                if flag & self.DESTINATION_FLAG:
                    obj.destination = decode(context)
                else:
                   obj.destination = None

                if flag & self.HEADERS_FLAG:
                    obj.headers = decode(context)
                else:
                    obj.headers = None

                if flag & self.MESSAGE_ID_FLAG:
                    obj.messageId = decode(context)
                else:
                    obj.messageId = None

                if flag & self.TIMESTAMP_FLAG:
                    obj.timestamp = decode(context)
                else:
                    obj.timestamp = None

                if flag & self.TIME_TO_LIVE_FLAG:
                    obj.timeToLive = decode(context)
                else:
                    obj.timeToLive = None

            if i == 1:
                if flag & self.CLIENT_ID_BYTES_FLAG:
                    clientIdBytes = decode(context)
                    obj.clientId = self._readUid(clientIdBytes)
                else:
                    if not hasattr(obj, 'clientId'):
                        obj.clientId = None

                if flag & self.MESSAGE_ID_BYTES_FLAG:
                    messageIdBytes = decode(context)
                    obj.messageId = self._readUid(messageIdBytes)
                else:
                    if not hasattr(obj, 'messageId'):
                        obj.messageId = None

class RemotingMessage(AbstractMessage):

    def __init__(self, body=None, clientId=None, destination=None,
        headers=None, timeToLive=None, timestamp=None, messageId=None,
        operation=None, source=None):

        AbstractMessage.__init__(self, body=body, clientId=clientId,
            destination=destination, headers=headers, timeToLive=timeToLive,
            timestamp=timestamp, messageId=messageId)

        self.operation = operation
        self.source = source

    def invoke(self, packet, msg):
        AbstractMessage.invoke(self, packet, msg)

        # Set connection object, so it is accessable in target.
        self.connection = packet.channel.getFlexConnection(self)

        target = packet.channel.channel_set.service_mapper.getTarget(self.destination, self.operation)        
        if target is None:
            raise FlexMessageError("Operation '%s' not found." % \
                remoting.Service.SEPARATOR.join((self.destination, self.operation)))

        if target.secure is True:
            if self.connection.authenticated is False:
                from amfast.remoting.channel import SecurityError
                raise SecurityError("Operation requires authentication.")

        msg.response_msg.body.body = target.invoke(packet, msg, self.body)

class_def.assign_attrs(RemotingMessage, 'flex.messaging.messages.RemotingMessage',
    ('body', 'clientId', 'destination', 'headers',
        'messageId', 'timestamp', 'timeToLive', 'source', 'operation'), True)

class AsyncMessage(AbstractMessage):

    SUBTOPIC_HEADER = 'DSSubtopic'

    def __init__(self, body=None, clientId=None, destination=None,
        headers=None, timeToLive=None, timestamp=None, messageId=None,
        correlationId=None):
 
               
        AbstractMessage.__init__(self, body=body, clientId=clientId,
            destination=destination, headers=headers, timeToLive=timeToLive,
            timestamp=timestamp, messageId=messageId) 

        if correlationId is not None:
            self.correlationId = correlationId

    def invoke(self, packet, msg):
        """Publish this message."""

        AbstractMessage.invoke(self, packet, msg)

        channel = packet.channel
        channel_set = channel.channel_set
        self.connection = channel.getFlexConnection(self)

        if channel_set.subscription_manager.secure is True:
            if self.connection.authenticated is False:
                from amfast.remoting.channel import SecurityError
                raise SecurityError("Operation requires authentication.")

        channel_set.publishMessage(self)

class_def.assign_attrs(AsyncMessage, 'flex.messaging.messages.AsyncMessage',
    ('body', 'clientId', 'destination', 'headers',
        'messageId', 'timestamp', 'timeToLive', 'correlationId'), True)

class AsyncSmallMsgDef(AbstractSmallMsgDef):
    """Decodes messages that were encoded using ISmallMessage."""

    CORRELATION_ID_FLAG = 0x01
    CORRELATION_ID_BYTES_FLAG = 0x02

    def readExternal(self, obj, context):
        AbstractSmallMsgDef.readExternal(self, obj, context)

        flags = self._readFlags(context)
        for i, flag in enumerate(flags):
            if i == 0:
                if flag & self.CORRELATION_ID_FLAG:
                    obj.correlationId = decode(context)
                else:
                    obj.correlationId = None

                if flag & self.CORRELATION_ID_BYTES_FLAG:
                    correlationIdBytes = decode(context)
                    obj.correlationId = self._readUid(correlationIdBytes)
                else:
                    if not hasattr(obj, 'correlationId'):
                        obj.correlationId = None

class CommandMessage(AsyncMessage):
    """A Flex CommandMessage. Operations are integers instead of strings.

    See Flex API docs for list of possible commands.
    """

    SUBSCRIBE_OPERATION = 0
    UNSUBSCRIBE_OPERATION = 1
    POLL_OPERATION = 2
    CLIENT_SYNC_OPERATION = 4
    CLIENT_PING_OPERATION = 5
    CLUSTER_REQUEST_OPERATION = 7
    LOGIN_OPERATION = 8
    LOGOUT_OPERATION = 9
    SUBSCRIPTION_INVALIDATE_OPERATION = 10
    MULTI_SUBSCRIBE_OPERATION = 11
    DISCONNECT_OPERATION = 12
    TRIGGER_CONNECT_OPERATION = 13

    ADD_SUBSCRIPTIONS = 'DSAddSub'
    CREDENTIALS_CHARSET_HEADER = 'DSCredentialsCharset'
    MAX_FREQUENCY_HEADER = 'DSMaxFrequency'
    MESSAGING_VERSION = 'DSMessagingVersion'
    NEEDS_CONFIG_HEADER = 'DSNeedsConfig'
    NO_OP_POLL_HEADER = 'DSNoOpPoll'
    POLL_WAIT_HEADER = 'DSPollWait'
    PRESERVE_DURABLE_HEADER = 'DSPreserveDurable'
    REMOVE_SUBSCRIPTIONS = 'DSRemSub'
    SELECTOR_HEADER = 'DSSelector'
    SUBTOPIC_SEPARATOR = '_;_'

    def __init__(self,  body=None, clientId=None, destination=None,
        headers=None, timeToLive=None, timestamp=None, messageId=None,
        correlationId=None, operation=1000):

        AsyncMessage.__init__(self, body=body, clientId=clientId,
            destination=destination, headers=headers, timeToLive=timeToLive,
            timestamp=timestamp, messageId=messageId, correlationId=correlationId)

        self.operation = operation

    def invoke(self, packet, msg):
        AbstractMessage.invoke(self, packet, msg)

        self.connection = packet.channel.getFlexConnection(self)

        target = packet.channel.channel_set.service_mapper.command_service.getTarget(self.operation)
        if target is None:
            raise FlexMessageError("Command '%s' not found." % self.operation)

        msg.response_msg.body.body = target.invoke(packet, msg, (self.body,))

    def getAcknowledgeClass(self):
        """Returns the correct class for the response message."""
        if self.operation == self.POLL_OPERATION:
            return CommandMessage

        return AsyncMessage.getAcknowledgeClass(self)

class_def.assign_attrs(CommandMessage, 'flex.messaging.messages.CommandMessage',
    ('body', 'clientId', 'destination', 'headers',
        'messageId', 'timestamp', 'timeToLive', 'correlationId',
        'operation'), True)

class CommandSmallMsgDef(AsyncSmallMsgDef):
    """Decodes messages that were encoded using ISmallMessage."""

    OPERATION_FLAG = 0x01

    def readExternal(self, obj, context):
        AsyncSmallMsgDef.readExternal(self, obj, context)

        flags = self._readFlags(context)
        for i, flag in enumerate(flags):
            if i == 0:
                if flag & self.OPERATION_FLAG:
                    obj.operation = decode(context)
                else:
                    obj.operation = None

class AcknowledgeMessage(AsyncMessage):
    """A response message sent back to the client."""

    ERROR_HINT_HEADER = 'DSErrorHint'

class_def.assign_attrs(AcknowledgeMessage, 'flex.messaging.messages.AcknowledgeMessage',
    ('body', 'clientId', 'destination', 'headers',
        'messageId', 'timestamp', 'timeToLive', 'correlationId'), True)

class ErrorMessage(AcknowledgeMessage):
    """A response message sent back to the client after a failure."""

    def __init__(self, body=None, clientId=None, destination=None,
        headers=None, timeToLive=None, timestamp=None, messageId=None,
        correlationId=None, exc=None, faultCode='', faultString='',
        faultDetail='', rootCause=None, extendedData=None):
        """exc must be a FaultError or None."""
       
        AcknowledgeMessage.__init__(self, body=body, clientId=clientId,
            destination=destination, headers=headers, timeToLive=timeToLive,
            timestamp=timestamp, messageId=messageId, correlationId=correlationId)

        if exc is not None:
            self.faultCode = exc.faultCode
            self.faultString = exc.faultString
            self.faultDetail = exc.faultDetail
            self.rootCause = exc.rootCause
            self.extendedData = exc.content
            self.body = exc
        else:
            self.faultCode = faultCode
            self.faultString = faultString
            self.faultDetail = faultDetail
            self.rootCause = rootCause
            self.extendedData = extendedData

class_def.assign_attrs(ErrorMessage, 'flex.messaging.messages.ErrorMessage', 
    ('body', 'clientId', 'destination', 'headers',
        'messageId', 'timestamp', 'timeToLive', 'correlationId', 'faultCode',
        'faultString', 'faultDetail', 'rootCause', 'extendedData'), True)

class StreamingMessage(CommandMessage):
    """A command message delivered over an HTTP streaming channel."""

    # streaming command key
    COMMAND_PARAM_NAME = 'command'

    # open a streaming connection
    OPEN_COMMAND = 'open'

    # close streaming connection
    CLOSE_COMMAND = 'close'

    # stream id
    ID_PARAM_NAME = 'streamId'

    # stream version
    VERSION_PARAM_NAME = 'version'

    # Bytes for encoding message
    CR_BYTE = 13
    LF_BYTE = 10
    NULL_BYTE = 0

    @classmethod
    def getDisconnectMsg(cls):
        msg = CommandMessage()
        msg.operation = CommandMessage.DISCONNECT_OPERATION
        return msg

    @classmethod
    def prepareMsg(cls, msg, endpoint):
        return cls.getMsgBytes(endpoint.encode(msg, amf3=True))

    @classmethod
    def getMsgBytes(cls, raw):
        """Add size information to raw AMF encoding for streaming."""
        byte_size = len(raw)
        hex_len = '%x' % byte_size # Turn length into a string of hex digits
        return ''.join((hex_len, chr(cls.CR_BYTE), chr(cls.LF_BYTE), raw)) # CR_BYTE marks end of size declaration, LF_BYTE marks beginning of data section.

    def parseArgs(self, args):
        if not hasattr(self, 'headers') or self.headers is None:
            self.headers = {}
        
        if self.FLEX_CLIENT_ID_HEADER in args:
            self.headers[self.FLEX_CLIENT_ID_HEADER] = args[self.FLEX_CLIENT_ID_HEADER][0]

        if self.COMMAND_PARAM_NAME in args:
            self.operation = args[self.COMMAND_PARAM_NAME][0]

    def parseParams(self, url_params):
        """Parses and sets attributes from URL parameters."""
        params = cgi.parse_qs(url_params, True)
        self.operation = params.get(self.COMMAND_PARAM_NAME)[0]

    def parseBody(self, body):
        if not hasattr(self, 'headers') or self.headers is None:
            self.headers = {}

        params = cgi.parse_qsl(body, True)
        for param in params:
            if param[0] == self.FLEX_CLIENT_ID_HEADER:
                self.headers[self.FLEX_CLIENT_ID_HEADER] = param[1]

    def acknowledge(self, *args, **kwargs):
        """Return a successful result message."""
        class_ = self.getAcknowledgeClass()
        response = class_()
        self._matchAcknowledge(response)
        return response

    def _matchAcknowledge(self, response, *args, **kwargs):
        """Syncs values between this message and it's response acknowledgement."""
        response.correlationId = self.operation



"""Channels that can be used with WSGI."""
import time
import threading

import amfast
from amfast import AmFastError
from amfast.remoting.channel import ChannelSet, HttpChannel, ChannelError
import amfast.remoting.flex_messages as messaging

class WsgiChannelSet(ChannelSet):
    def __call__(self, environ, start_response):
        channel_name = environ['PATH_INFO'][1:]
        channel = self.getChannel(channel_name)
        return channel(environ, start_response)

class WsgiChannel(HttpChannel):
    """WSGI app channel."""

    def __init__(self, *args, **kwargs):

        if len(args) > 3:
            wait_interval = args[3]
        elif 'wait_interval' in kwargs:
            wait_interval = kwargs['wait_interval']
        else:
            wait_interval = 0

        if wait_interval < 0:
            # The only reliable way to detect
            # when a client has disconnected with
            # WSGI is that the 'write' function will fail.
            #
            # With long-polling, nothing is written
            # until a message is published to a client,
            # so we are unable to detect if we are
            # waiting for a message for a disconnected client.
            #
            # wait_interval must be non-negative to avoid
            # zombie threads.
            raise ChannelError('wait_interval < 0 is not supported by WsgiChannel')
            
        HttpChannel.__init__(self, *args, **kwargs)

    def __call__(self, environ, start_response):
        if environ['REQUEST_METHOD'] != 'POST':
            return self.badMethod(start_response)

        len_str = 'CONTENT_LENGTH'
        raw_request = environ['wsgi.input'].read(int(environ[len_str]))

        try:
            request_packet = self.decode(raw_request)
        except AmFastError, exc:
            return self.badRequest(start_response, self.getBadEncodingMsg())
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            return self.badServer(start_response, self.getBadServerMsg())

        try:
            content = self.invoke(request_packet)
            response = self.encode(content)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            return self.badServer(start_response, self.getBadServerMsg())

        return self.getResponse(start_response, response)

    def getResponse(self, start_response, response):
        start_response('200 OK', [
            ('Content-Type', self.CONTENT_TYPE),
            ('Content-Length', str(len(response)))
        ])

        return [response]

    def badMethod(self, start_response):
        response = self.getBadMethodMsg()

        start_response('405 Method Not Allowed', [
            ('Content-Type', 'text/plain'),
            ('Content-Length', str(len(response)))
        ])

        return [response]

    def badRequest(self, start_response, response):
        start_response('400 Bad Request', [
            ('Content-Type', 'text/plain'),
            ('Content-Length', str(len(response)))
        ])

        return [response]

    def badPage(self, start_response, response):
        start_response('404 Not Found', [
            ('Content-Type', 'text/plain'),
            ('Content-Length', str(len(response)))
        ])

        return [response]

    def badServer(self, start_response, response):
        start_response('500 Internal Server Error', [
            ('Content-Type', 'text/plain'),
            ('Content-Length', str(len(response)))
        ])

        return [response]

class StreamingWsgiChannel(WsgiChannel):
    """WsgiChannel that opens a persistent connection with the client to serve messages."""

    def __init__(self, name, max_connections=-1, endpoint=None, wait_interval=0, heart_interval=30):
        WsgiChannel.__init__(self, name, max_connections=max_connections,
            endpoint=endpoint, wait_interval=wait_interval)

        self.heart_interval = heart_interval

    def __call__(self, environ, start_response):
        if environ['CONTENT_TYPE'] == self.CONTENT_TYPE:
            # Regular AMF message
            return WsgiChannel.__call__(self, environ, start_response)
        
        # Create streaming message command
        try:
            msg = messaging.StreamingMessage()
            msg.parseParams(environ['QUERY_STRING'])

            body = environ['wsgi.input'].read(int(environ['CONTENT_LENGTH']))
            msg.parseBody(body)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            return self.badServer(start_response, self.getBadServerMsg())

        if msg.operation == msg.OPEN_COMMAND:
            return self.startStream(environ, start_response, msg)

        if msg.operation == msg.CLOSE_COMMAND:
            return self.stopStream(msg)

        return self.badRequest(start_response, self.getBadRequestMsg('Streaming operation unknown: %s' % msg.operation))

    def startStream(self, environ, start_response, msg):
        """Start streaming response."""

        try: 
            connection = self.channel_set.connection_manager.getConnection(msg.headers.get(msg.FLEX_CLIENT_ID_HEADER))
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            return self.badServer(start_response, self.getBadServerMsg())

        write = start_response('200 OK', [
            ('Content-Type', self.CONTENT_TYPE)
        ])

        try:
            # Send acknowledge message
            response = msg.acknowledge()
            response.body = connection.id

            try:
                bytes = messaging.StreamingMessage.prepareMsg(response, self.endpoint)
                write(bytes)

                write(chr(messaging.StreamingMessage.NULL_BYTE) * self.KICKSTART_BYTES)

            except (KeyboardInterrupt, SystemExit):
                raise
            except Exception, exc:
                amfast.log_exc(exc)
                return []

            # Start heart beat
            timer = threading.Timer(self.heart_interval, self.beat, (connection, ))
            timer.daemon = True
            timer.start()

            # Wait for new messages.
            event = threading.Event()
            connection.setNotifyFunc(event.set)
            poll_secs = float(self.poll_interval) / 1000
            while True:

                if connection.connected is False:
                    # Connection is no longer active
                    msg = messaging.StreamingMessage.getDisconnectMsg()
                    try:
                        write(messaging.StreamingMessage.prepareMsg(msg, self.endpoint))
                    except:
                        # Client may have already disconnected
                        pass
                    # Stop stream
                    return []

                if self.channel_set.notify_connections is True: 
                    # Block until notification of new message
                    event.wait()
                else:
                    # Block until poll_interval is reached
                    event.wait(poll_secs)

                # Message has been published,
                # or it's time for a heart beat

                # Remove notify_func so that
                # New messages don't trigger event.
                connection.unSetNotifyFunc()

                msgs = self.channel_set.subscription_manager.pollConnection(connection)
                if len(msgs) > 0:
                    while len(msgs) > 0:
                        # Dispatch all messages to client
                        for msg in msgs:
                            try:
                                bytes = messaging.StreamingMessage.prepareMsg(msg, self.endpoint)
                            except (KeyboardInterrupt, SystemExit):
                                raise
                            except Exception, exc:
                                amfast.log_exc(exc)
                                self.channel_set.disconnect(connection)
                                break

                            try:
                                write(bytes)
                            except (KeyboardInterrupt, SystemExit):
                                raise
                            except:
                                # Client has disconnected
                                self.channel_set.disconnect(connection)
                                return []

                        msgs = self.channel_set.subscription_manager.pollConnection(connection)
                else:
                    # Send heart beat
                    try:
                        write(chr(messaging.StreamingMessage.NULL_BYTE))
                    except (KeyboardInterrupt, SystemExit):
                        raise
                    except:
                        # Client has disconnected
                        self.channel_set.disconnect(connection)
                        return []

                # Create new event to trigger new messages or heart beats
                event = threading.Event()
                connection.setNotifyFunc(event.set)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            self.channel_set.disconnect(connection)
            return []

    def beat(self, connection):
        """Send a heart beat."""
        if hasattr(connection, "notify_func") and connection.notify_func is not None:
            connection.notify_func()
        else:
            return

        # Create timer for next beat
        timer = threading.Timer(self.heart_interval, self.beat, (connection, ))
        timer.daemon = True
        timer.start()

    def stopStream(self, msg):
        """Stop a streaming connection."""
        connection = self.channel_set.connection_manager.getConnection(msg.headers.get(msg.FLEX_CLIENT_ID_HEADER))
        connection.disconnect()
        if hasattr(connection, "notify_func") and connection.notify_func is not None:
            connection.notify_func()



from amfast.decode import decode, decode_packet
from amfast.context import DecoderContext
from amfast.class_def import ClassDefMapper

class Decoder(object):
    """A wrapper class for convenient access to amfast.decode.decode.

    Decoder
    ========
     * amf3 - bool - True to decode as AMF3.
     * class_def_mapper - amfast.class_def.ClassDefMapper - The object that retrieves ClassDef objects.
    """ 

    def __init__(self, amf3=False, class_def_mapper=None):

        self.amf3 = amf3

        if class_def_mapper is None:
            class_def_mapper = ClassDefMapper()
        self.class_def_mapper = class_def_mapper

    def _getContext(self, input, amf3=None):
        if amf3 is None:
            amf3 = self.amf3
        return DecoderContext(input, amf3=amf3, class_def_mapper=self.class_def_mapper)

    def decode(self, val, amf3=None):
        """Decode a string or file-like-object from AMF."""
        return decode(self._getContext(val, amf3))

    def decode_packet(self, val, amf3=None):
        """Decode a string or file-like-object representing an AMF packet."""
        return decode_packet(self._getContext(val, amf3))



"""Serialize/De-serialize Python objects with AMF3."""
import sys
import traceback
import calendar
from datetime import datetime
import logging

from amfast import version
version = version.Version('amfast', 0, 5, 2)
__version__ = version.short()

class AmFastError(Exception):
    """Base exception for this package."""

    # To get around 2.6 deprecation warning
    def _get_message(self):
        return self._message
    def _set_message(self, message):
        self._message = message
    message = property(_get_message, _set_message)

def epoch_from_date(date):
    """Returns epoch milliseconds."""
    return long(calendar.timegm(date.timetuple()) * 1000)

def date_from_epoch(epoch_secs):
    """Returns datetime."""
    return datetime.utcfromtimestamp(epoch_secs)

def format_byte_string(byte_string):
    """Get a human readable description of a byte string."""
    bytes = []
    for i, x in enumerate(byte_string):
        val = ord(x)
        char = ''
        if val > 31 and val < 127:
            char = "%s" % x
        bytes.append("%d: %d-%02X-'%s'" % (i, val, val, char))
    return ' '.join(bytes)

# --- setup module level logging --- #

class NullHandler(logging.Handler):
    """A logging handler that does nothing, so that end-users
    do not receive 'No handlers could be found...' warning.
 
    Yes, this is the way the Python docs recommend solving this problem.
    """
    def emit(self, record):
        pass

log_debug = False # True to log verbose debug strings.
log_raw = False # True to log raw AMF byte strings.
logged_attr = '_amfast_logged' # Add to exceptions to indicate that it has been logged.
logger = logging.getLogger('AmFast')
logger.addHandler(NullHandler())
logger.setLevel(logging.DEBUG)

month_names = [
    'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
    'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']

def get_log_timestamp():
    dt = datetime.now()
    return '[%s/%s/%s:%s:%s:%s]' % (dt.day, month_names[dt.month - 1],
        dt.year, dt.hour, dt.minute, dt.second)

def log_exc(e):
    """Format an exception for logging."""
    if hasattr(e, logged_attr):
        return
    else:
        setattr(e, logged_attr, True)

    error_type, error_value, trbk = sys.exc_info()
    tb_list = traceback.format_tb(trbk)
    
    msg = [get_log_timestamp() + " EXCEPTION"]
    msg.append("# ---- EXCEPTION DESCRIPTION BEGIN ---- #")
    
    msg.append("# ---- Type ---- #\n%s\n# ---- Detail ---- #\n%s" % \
        (error_type.__name__, error_value))
    msg.append("# ---- Traceback ---- #")
    msg.append("-\n".join(tb_list))
    msg.append("# ---- EXCEPTION DESCRIPTION END ---- #")
    logger.error("\n".join(msg))

# --- Setup threading implementation --- #

try:
    import threading
    mutex_cls = threading.RLock
    use_dummy_threading = False
except ImportError:
    import dummy_threading
    mutex_cls = dummy_threading.RLock
    use_dummy_threading = True

    if log_debug:
        logger.debug("AmFast is using dummy_threading module.")



import time

import amfast
import flex_messages as messaging

class ConnectionError(amfast.AmFastError):
    pass

class Connection(object):
    """A client connection to a Channel.

    This class acts like a session.
    Unique to a single flash client.

    attributes (read-only)
    =======================
     * manager - ConnectionManager, the class that manages connection persistence.
     * channel_name - string, name of the channel connection is connected to.
     * id - string, Flash client's ID.
     * timeout - int, timeout in milliseconds
     * connected - boolean, True if connection is connected
     * last_active - float, epoch timestamp when connection was last accessed.
     * last_polled - float, epocj timestamp when connection last polled for messsages.
     * authenticated - boolean, True if connection has been authenticated with RemoteObject style authentication.
     * flex_user - string, username if 'authenticated' is True.

    """

    # Keeps track of notification functions
    # by their IDs, 
    # so they can be serialized
    # when the connection is saved.
    _notifications = {}

    @classmethod
    def _setNotifyFunc(cls, func):
        func_id = id(func)
        cls._notifications[func_id] = func
        return func_id

    @classmethod
    def _getNotifyFuncById(cls, func_id):
        return cls._notifications[func_id]

    @classmethod
    def _delNotifyFunc(cls, func_id):
        del cls._notifications[func_id]

    def __init__(self, manager, channel_name, id, timeout=1800000):
        # These attributes should not
        # changed during the life of a connection.
        #
        # All other attributes that may
        # change during the life of a connection
        # should be accessed through properties
        # that call methods in the connection_manager.
        self._manager = manager
        self._channel_name = channel_name
        self._id = id
        self._timeout = timeout

    # --- read-only properties --- #
    def _getManager(self):
        return self._manager
    manager = property(_getManager)

    def _getChannelName(self):
        return self._channel_name
    channel_name = property(_getChannelName)

    def _getId(self):
        return self._id
    id = property(_getId)

    def _getTimeout(self):
        return self._timeout
    timeout = property(_getTimeout)

    # --- proxied properties --- #

    def _getConnected(self):
        return self._manager.getConnected(self)
    connected = property(_getConnected)

    def _getLastActive(self):
        return self._manager.getLastActive(self)
    last_active = property(_getLastActive)

    def _getLastPolled(self):
        return self._manager.getLastPolled(self)
    last_polled = property(_getLastPolled)

    def _getAuthenticated(self):
        return self._manager.getAuthenticated(self)
    authenticated = property(_getAuthenticated)

    def _getFlexUser(self):
        return self._manager.getFlexUser(self)
    flex_user = property(_getFlexUser)

    def _getNotifyFunc(self):
        return self._manager.getNotifyFunc(self)
    notify_func = property(_getNotifyFunc)

    # --- proxied methods --- #

    def touch(self):
        """Update last_active."""
        self._manager.touchConnection(self)

    def touchPolled(self):
        """Update last_polled."""
        self._manager.touchPolled(self)

    def softTouchPolled(self):
        """Update last_polled without persisting value.
  
        Useful when ChannelSet calls _pollForMessage.
        """
        self._manager.softTouchPolled(self)

    def connect(self):
        """Set connected=True."""
        self._manager.connectConnection(self)

    def disconnect(self):
        """Set connected=False."""
        self._manager.disconnectConnection(self)

    def delete(self):
        """Delete connection."""
        self._manager.deleteConnection(self)

    def authenticate(self, user):
        """Set authenticated = True"""
        self._manager.authenticateConnection(self, user)

    def unAuthenticate(self):
        """Set authenticated = False"""
        self._manager.unAuthenticateConnection(self)

    def setNotifyFunc(self, func):
        self._manager.setNotifyFunc(self, func)

    def unSetNotifyFunc(self):
        self._manager.unSetNotifyFunc(self)

    def getSessionAttr(self, name):
        """Get a session attribute."""
        return self._manager.getConnectionSessionAttr(self, name)

    def setSessionAttr(self, name, val):
        """Set a session attribute."""
        self._manager.setConnectionSessionAttr(self, name, val)

    def delSessionAttr(self, name):
        """Del a session attribute."""
        self._manager.delConnectionSessionAttr(self, name)

    # --- instance methods --- #

    def personalizeMessage(self, client_id, msg):
        """Return a copy of the message with client_id set."""

        if hasattr(msg, 'headers'):
            headers = msg.headers
        else:
            headers = None

        return msg.__class__(headers=headers, body=msg.body,
            timeToLive=msg.timeToLive, clientId=client_id,
            destination=msg.destination, timestamp=msg.timestamp)



"""Channels that can be used with the CherryPy framework."""
import threading

import cherrypy
import cherrypy.process.plugins as cp_plugins

import amfast
import amfast.remoting.flex_messages as messaging
from amfast.remoting.channel import ChannelSet, HttpChannel, ChannelError

def amfhook():
    """Checks for POST, and stops cherrypy from processing the body."""

    cherrypy.request.process_request_body = False
    cherrypy.request.show_tracebacks = False

    if cherrypy.request.method != 'POST':
        raise cherrypy.HTTPError(405, "405 Method Not Allowed\n\nAMF request must use 'POST' method.");
cherrypy.tools.amfhook = cherrypy.Tool('before_request_body', amfhook, priority=0)

class CherryPyChannelSet(ChannelSet):
    """A ChannelSet for use with CherryPy."""

    def __init__(self, *args, **kwargs):
        self.clean_scheduled = False
        ChannelSet.__init__(self, *args, **kwargs)

    def scheduleClean(self):
        """Overridden to use CherryPy's Monitor functionality."""
        if self.clean_scheduled is False:
            cleaner = cp_plugins.Monitor(cherrypy.engine, self.clean, self.clean_freq)
            cleaner.name = "ConnectionCleaner"
            cleaner.subscribe()
            self.clean_scheduled = True

    def mapChannel(self, channel):
        """Overridden so that channel is added as an attribute."""
        if hasattr(self, channel.name):
            raise ChannelError("Reserved attribute name '%s' cannot be used as a channel name." % channel.name)

        ChannelSet.mapChannel(self, channel)

        setattr(self, channel.name, channel.__call__)

class CherryPyChannel(HttpChannel):
    """An AMF messaging channel that can be used with CherryPy HTTP framework.

    Instantiate a CherryPyChannel object and
    mount the __call__ method to the URL where
    AMF messaging should be available from. 

    You can also use a WsgiChannel instance by
    grafting it to the CherryPy tree with
    the command cherrypy.tree.graft.

    Using WsgiChannel will be more efficient, but you
    won't have access to any of CherryPy's built-in tools
    such as cookie-based sessions.

    """

    @cherrypy.expose
    @cherrypy.tools.amfhook()
    def __call__(self):
        try:
            c_len = int(cherrypy.request.headers['Content-Length'])
            raw_request = cherrypy.request.rfile.read(c_len)
        except KeyError:
            raw_request = cherrypy.request.rfile

        response = self.invoke(self.decode(raw_request))
        cherrypy.response.headers['Content-Type'] = self.CONTENT_TYPE
        return self.encode(response)

class StreamingCherryPyChannel(CherryPyChannel):
    """Allows HTTP streaming."""

    def __init__(self, name, max_connections=-1, endpoint=None,
        wait_interval=0, heart_interval=30):
        CherryPyChannel.__init__(self, name, max_connections=max_connections,
            endpoint=endpoint, wait_interval=wait_interval)

        self.heart_interval = heart_interval

    @cherrypy.expose
    @cherrypy.tools.amfhook()
    def __call__(self, command=None, version=None):
        if cherrypy.request.headers['Content-Type'] == self.CONTENT_TYPE:
            # Regular AMF message
            return CherryPyChannel.__call__(self)

        # Create streaming message command
        cherrypy.response.headers['Content-Type'] = self.CONTENT_TYPE
        cherrypy.response.stream = True
        try:
            msg = messaging.StreamingMessage()
            msg.parseParams(cherrypy.request.query_string)

            c_len = int(cherrypy.request.headers['Content-Length'])
            body = cherrypy.request.rfile.read(c_len)
            msg.parseBody(body)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            raise ChannelError("AMF server error.")

        if msg.operation == msg.OPEN_COMMAND:
            return self.startStream(msg)

        if msg.operation == msg.CLOSE_COMMAND:
            return self.stopStream(msg)

        raise ChannelError('Http streaming operation unknown: %s' % msg.operation)

    def startStream(self, msg):
        """Returns an iterator for streaming."""

        try:
            connection = self.channel_set.connection_manager.getConnection(msg.headers.get(msg.FLEX_CLIENT_ID_HEADER))
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            raise ChannelError('Http streaming operation unknown: %s' % msg.operation)

        cherrypy.response.headers['Content-Type'] = self.CONTENT_TYPE

        try:
            # Start heart beat
            timer = threading.Timer(self.heart_interval, self.beat, (connection, ))
            timer.daemon = True
            timer.start()

            # Wait for new messages.
            inited = False
            event = threading.Event()
            connection.setNotifyFunc(event.set)
            poll_secs = float(self.poll_interval) / 1000
            while True:

                if connection.connected is False:
                    # Connection is no longer active
                    msg = messaging.StreamingMessage.getDisconnectMsg()
                    try:
                        yield messaging.StreamingMessage.prepareMsg(msg, self.endpoint)
                    finally:
                        # Client may have already disconnected
                        return

                if inited is False:
                    # Send acknowledge message
                    response = msg.acknowledge()
                    response.body = connection.id

                    bytes = messaging.StreamingMessage.prepareMsg(response, self.endpoint)
                    inited = True

                    bytes += chr(messaging.StreamingMessage.NULL_BYTE) * self.KICKSTART_BYTES
                    yield bytes
 
                if self.channel_set.notify_connections is True:
                    # Block until notification of new message
                    event.wait()
                else:
                    # Block until poll_interval is reached
                    event.wait(poll_secs)

                # Message has been published,
                # or it's time for a heart beat

                # Remove notify_func so that
                # New messages don't trigger event.
                connection.unSetNotifyFunc()

                msgs = self.channel_set.subscription_manager.pollConnection(connection)
                if len(msgs) > 0:
                    while len(msgs) > 0:
                        # Dispatch all messages to client
                        for msg in msgs:
                            try:
                                bytes = messaging.StreamingMessage.prepareMsg(msg, self.endpoint)
                            except (KeyboardInterrupt, SystemExit):
                                raise
                            except Exception, exc:
                                amfast.log_exc(exc)
                                self.channel_set.disconnect(connection)
                                break

                            try:
                                yield bytes
                            except (KeyboardInterrupt, SystemExit):
                                raise
                            except:
                                # Client has disconnected
                                self.channel_set.disconnect(connection)
                                return

                        msgs = self.channel_set.subscription_manager.pollConnection(connection)
                else:
                    # Send heart beat
                    try:
                        yield chr(messaging.StreamingMessage.NULL_BYTE)
                    except (KeyboardInterrupt, SystemExit):
                        raise
                    except:
                        # Client has disconnected
                        self.channel_set.disconnect(connection)
                        return

                # Create new event to trigger new messages or heart beats
                event = threading.Event()
                connection.setNotifyFunc(event.set)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, exc:
            amfast.log_exc(exc)
            self.channel_set.disconnect(connection)
            return

    def stopStream(self, msg):
        """Stop a streaming connection."""

        connection = self.channel_set.getConnection(msg.headers.get(msg.FLEX_CLIENT_ID_HEADER))
        connection.disconnect()
        if hasattr(connection, "notify_func") and connection.notify_func is not None:
            connection.notify_func()

    def beat(self, connection):
        """Send a heart beat."""

        if hasattr(connection, "notify_func") and connection.notify_func is not None:
            connection.notify_func()
        else:
            return

        # Create timer for next beat
        timer = threading.Timer(self.heart_interval, self.beat, (connection, ))
        timer.daemon = True
        timer.start()



from amfast.encode import encode, encode_packet
from amfast.context import EncoderContext
from amfast.class_def import ClassDefMapper

class Encoder(object):
    """A wrapper class for convenient access to amfast.encode.encode.

    Encoder
    ========
     * amf3 - bool - True to encode as AMF3.
     * use_collections - bool - True to encode lists and tuples as ArrayCollections.
     * use_proxies - bool - True to encode dicts as ObjectProxies.
     * use_references - bool - True to encode multiply occuring objects by reference.
     * use_legacy_xml - bool - True to XML as XMLDocument instead of e4x.
     * include_private - bool - True to encode attributes starting with '_'.
     * class_def_mapper - amfast.class_def.ClassDefMapper - The object that retrieves ClassDef objects.
     * buffer - file-like-object - Output buffer. Set to None to output to a string.

    """ 

    def __init__(self, amf3=False, use_collections=False, use_proxies=False,
        use_references=True, use_legacy_xml=False, include_private=False,
        class_def_mapper=None, buffer=None):

        self.amf3 = amf3
        self.use_collections = use_collections
        self.use_proxies = use_proxies
        self.use_references = use_references
        self.use_legacy_xml = use_legacy_xml
        self.include_private = include_private

        if class_def_mapper is None:
            class_def_mapper = ClassDefMapper()
        self.class_def_mapper = class_def_mapper

        self.buffer = buffer

    def _getContext(self, amf3=None):
        if amf3 is None:
            amf3 = self.amf3

        kwargs = {
            'amf3': amf3,
            'use_collections': self.use_collections,
            'use_proxies': self.use_proxies,
            'use_references': self.use_references,
            'use_legacy_xml': self.use_legacy_xml,
            'include_private': self.include_private,
            'class_def_mapper': self.class_def_mapper
        }
 
        if self.buffer is not None:
            kwargs['buffer'] = self.buffer

        return EncoderContext(**kwargs);

    def encode(self, val, amf3=None):
        """Encode a value to AMF."""
        return encode(val, self._getContext(amf3))

    def encode_packet(self, val, amf3=None):
        """Encode an AMF packet."""
        return encode_packet(val, self._getContext(amf3))



import time
import cPickle as pickle

import sqlalchemy as sa
from sqlalchemy.sql import func, and_

if sa.__version__.startswith('0.5'):
    BINARY_TYPE = sa.Binary
elif sa.__version__.startswith('0.6'):
    BINARY_TYPE = sa.LargeBinary

from subscription_manager import Subscription, SubscriptionManager
import flex_messages as messaging

class SaSubscriptionManager(SubscriptionManager):
    """Manages subscriptions in a database, uses SqlAlchemy to talk to the DB."""

    def __init__(self, engine, metadata, secure=False, ttl=30000, table_prefix=''):
        SubscriptionManager.__init__(self, secure=secure, ttl=ttl)

        self.engine = engine
        self.metadata = metadata
        self.table_prefix = table_prefix and "%s_" % table_prefix.rstrip('_') or table_prefix
        self.mapTables()

    def reset(self):
        db = self.getDb()
        db.execute(self.subscriptions.delete())
        db.execute(self.messages.delete())
        db.close()

    def mapTables(self):
        self.subscriptions = sa.Table('%ssubscriptions' % self.table_prefix, self.metadata,
            sa.Column('connection_id', sa.String(36), primary_key=True),
            sa.Column('client_id', sa.String(36), primary_key=True),
            sa.Column('topic', sa.String(128), primary_key=True)
        )

        self.messages = sa.Table('%smessages' % self.table_prefix, self.metadata,
            sa.Column('id', sa.Integer, primary_key=True),
            sa.Column('topic', sa.String(256), index=True),
            sa.Column('clientId', sa.String(128), nullable=True),
            sa.Column('messageId', sa.String(128), nullable=True),
            sa.Column('correlationId', sa.String(128), nullable=True),
            sa.Column('destination', sa.String(128), nullable=True),
            sa.Column('timestamp', sa.Float(), nullable=True),
            sa.Column('timeToLive', sa.Float(), nullable=True),
            sa.Column('headers', BINARY_TYPE(), nullable=True),
            sa.Column('body', BINARY_TYPE(), nullable=False)
        )

    def createTables(self):
       db = self.getDb()
       self.subscriptions.create(db, checkfirst=True)
       self.messages.create(db, checkfirst=True)
       db.close()

    def getDb(self):
        return self.engine.connect()

    def subscribe(self, connection_id, client_id, topic, sub_topic=None, selector=None):
        """Subscribe a client to a topic.

        arguments
        ==========
         * connection_id - string, id of Flash client that is subscribing.
         * client_id - string, id of messaging client that is subscribing.
         * topic - string, Topic to subscribe to.
         * sub_topic - string, Sub-Topic to subscribe to. Default = None.
        """

        topic = self.getTopicKey(topic, sub_topic)

        ins = self.subscriptions.insert().values(
            connection_id=connection_id,
            client_id=client_id,
            topic=topic
        )

        db = self.getDb()
        db.execute(ins)
        db.close()

    def unSubscribe(self, connection_id, client_id, topic, sub_topic=None):
        """Un-Subscribe a client from a topic.

        arguments
        ==========
         * connection_id - string, id of Flash client that is subscribing.
         * client_id - string, id of messaging client that is subscribing.
         * topic - string, Topic to un-subscribe from.
         * sub_topic - string, Sub-Topic to un-subscribe from. Default = None.
        """

        topic = self.getTopicKey(topic, sub_topic)

        d = self.subscriptions.delete().\
            where(and_(self.subscriptions.c.connection_id==connection_id,
                self.subscriptions.c.client_id==client_id,
                self.subscriptions.c.topic==topic))
        db = self.getDb()
        db.execute(d)
        db.close()

    def deleteConnection(self, connection):
        """Remove all subscriptions for this connection.

        arguments
        ==========
         * connection_id - string, id of Flash client that is subscribing.
        """

        d = self.subscriptions.delete().\
            where(self.subscriptions.c.connection_id==connection.id)
        db = self.getDb()
        db.execute(d)
        db.close()

    def iterSubscribers(self, topic, sub_topic=None):
        """Iterate through Flash client ids subscribed to a specific topic."""

        topic = self.getTopicKey(topic, sub_topic)

        s = sa.select([self.subscriptions.c.connection_id],
            self.subscriptions.c.topic==topic, distinct=True)
        db = self.getDb()
        results = db.execute(s)
        for row in results:
            yield row[self.subscriptions.c.connection_id]

    def iterConnectionSubscriptions(self, connection):
        """Iterate through all Subscriptions that belong to a specific connection."""

        s = sa.select([self.subscriptions.c.connection_id,
            self.subscriptions.c.client_id, self.subscriptions.c.topic],
            self.subscriptions.c.connection_id==connection.id)
        db = self.getDb()
        results = db.execute(s)
        for row in results:
            yield Subscription(row[self.subscriptions.c.connection_id],
                row[self.subscriptions.c.client_id], row[self.subscriptions.c.topic])

    def persistMessage(self, msg):
        """Store a message."""

        if hasattr(msg, 'headers') and (msg.headers is not None):
            enc_headers = pickle.dumps(msg.headers)
        else:
            enc_headers = None

        if hasattr(msg, 'correlationId'):
            correlation_id = msg.correlationId
        else:
            correlation_id = None

        ins = self.messages.insert().values(
            topic=self.getMessageTopicKey(msg),
            clientId=msg.clientId,
            messageId=msg.messageId,
            correlationId=correlation_id,
            destination=msg.destination,
            timestamp=msg.timestamp,
            timeToLive=msg.timeToLive,
            headers=enc_headers,
            body=pickle.dumps(msg.body)
        )

        db = self.getDb()
        db.execute(ins)
        db.close()

    def deleteExpiredMessages(self, cutoff_time):
        """Deletes expired messages."""
        
	d = self.messages.delete().\
            where(self.messages.c.timestamp + self.messages.c.timeToLive < cutoff_time)

        db = self.getDb()
        db.execute(d)
        db.close()

    def pollMessages(self, topic, cutoff_time, current_time):
        """Retrieves all queued messages, and discards expired messages.

        arguments:
        ===========
         * topic - string, Topic to find messages for.
         * cutoff_time - float, epoch time, only messages published
             after this time will be returned.
         * current_time - float, epoch time, used to determine if a
             message is expired.
        """

        # Poll for new messages
        s = sa.select((self.messages,),
                and_(self.messages.c.topic == topic,
                    self.messages.c.timestamp > cutoff_time)).\
            order_by(self.messages.c.timestamp)

        db = self.getDb()
        results = db.execute(s)
        for row in results:
             if row['headers'] is None:
                 headers = None
             else:
                 headers = pickle.loads(str(row['headers']))

             yield messaging.AsyncMessage(body=pickle.loads(str(row['body'])),
                 clientId=row['clientId'], destination=row['destination'],
                 headers=headers, timeToLive=row['timeToLive'],
                 timestamp=row['timestamp'], messageId=row['messageId'])

        db.close()



"""Tests to be run from Google App Engine."""
import unittest

import connection_test
import subscription_test

def suite():
    return unittest.TestSuite((
        unittest.TestLoader().loadTestsFromTestCase(connection_test.GaeTestCase),
        unittest.TestLoader().loadTestsFromTestCase(subscription_test.GaeTestCase),
        unittest.TestLoader().loadTestsFromTestCase(connection_test.MemcacheTestCase),
        unittest.TestLoader().loadTestsFromTestCase(subscription_test.MemcacheTestCase)
    ))
  



# -*- coding: utf-8 -*-

"""
Taken from Twisted package.
"""

"""
Copyright (c) 2001-2010
Allen Short
Andy Gayton
Andrew Bennetts
Antoine Pitrou
Apple Computer, Inc.
Benjamin Bruheim
Bob Ippolito
Canonical Limited
Christopher Armstrong
David Reid
Donovan Preston
Eric Mangold
Eyal Lotem
Itamar Shtull-Trauring
James Knight
Jason A. Mobarak
Jean-Paul Calderone
Jessica McKellar
Jonathan Jacobs
Jonathan Lange
Jonathan D. Simms
JÃ¼rgen Hermann
Kevin Horn
Kevin Turner
Mary Gardiner
Matthew Lefkowitz
Massachusetts Institute of Technology
Moshe Zadka
Paul Swartz
Pavel Pergamenshchik
Ralph Meijer
Sean Riley
Software Freedom Conservancy
Travis B. Hartwell
Thijs Triemstra
Thomas Herve
Timothy Allen

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import sys, os


class _inf(object):
    """
    An object that is bigger than all other objects.
    """
    def __cmp__(self, other):
        """
        @param other: Another object.
        @type other: any

        @return: 0 if other is inf, 1 otherwise.
        @rtype: C{int}
        """
        if other is _inf:
            return 0
        return 1

_inf = _inf()


class IncomparableVersions(TypeError):
    """
    Two versions could not be compared.
    """

class Version(object):
    """
    An object that represents a three-part version number.

    If running from an svn checkout, include the revision number in
    the version string.
    """
    def __init__(self, package, major, minor, micro, prerelease=None):
        """
        @param package: Name of the package that this is a version of.
        @type package: C{str}
        @param major: The major version number.
        @type major: C{int}
        @param minor: The minor version number.
        @type minor: C{int}
        @param micro: The micro version number.
        @type micro: C{int}
        @param prerelease: The prerelease number.
        @type prerelease: C{int}
        """
        self.package = package
        self.major = major
        self.minor = minor
        self.micro = micro
        self.prerelease = prerelease


    def short(self):
        """
        Return a string in canonical short version format,
        <major>.<minor>.<micro>[+rSVNVer].
        """
        s = self.base()
        svnver = self._getSVNVersion()
        if svnver:
            s += '+r' + str(svnver)
        return s


    def base(self):
        """
        Like L{short}, but without the +rSVNVer.
        """
        if self.prerelease is None:
            pre = ""
        else:
            pre = "pre%s" % (self.prerelease,)
        return '%d.%d.%d%s' % (self.major,
                               self.minor,
                               self.micro,
                               pre)


    def __repr__(self):
        svnver = self._formatSVNVersion()
        if svnver:
            svnver = '  #' + svnver
        if self.prerelease is None:
            prerelease = ""
        else:
            prerelease = ", prerelease=%r" % (self.prerelease,)
        return '%s(%r, %d, %d, %d%s)%s' % (
            self.__class__.__name__,
            self.package,
            self.major,
            self.minor,
            self.micro,
            prerelease,
            svnver)


    def __str__(self):
        return '[%s, version %s]' % (
            self.package,
            self.short())


    def __cmp__(self, other):
        """
        Compare two versions, considering major versions, minor versions, micro
        versions, then prereleases.

        A version with a prerelease is always less than a version without a
        prerelease. If both versions have prereleases, they will be included in
        the comparison.

        @param other: Another version.
        @type other: L{Version}

        @return: NotImplemented when the other object is not a Version, or one
            of -1, 0, or 1.

        @raise IncomparableVersions: when the package names of the versions
            differ.
        """
        if not isinstance(other, self.__class__):
            return NotImplemented
        if self.package != other.package:
            raise IncomparableVersions("%r != %r"
                                       % (self.package, other.package))

        if self.prerelease is None:
            prerelease = _inf
        else:
            prerelease = self.prerelease

        if other.prerelease is None:
            otherpre = _inf
        else:
            otherpre = other.prerelease

        x = cmp((self.major,
                    self.minor,
                    self.micro,
                    prerelease),
                   (other.major,
                    other.minor,
                    other.micro,
                    otherpre))
        return x


    def _parseSVNEntries_4(self, entriesFile):
        """
        Given a readable file object which represents a .svn/entries file in
        format version 4, return the revision as a string.  We do this by
        reading first XML element in the document that has a 'revision'
        attribute.
        """
        from xml.dom.minidom import parse
        doc = parse(entriesFile).documentElement
        for node in doc.childNodes:
            if hasattr(node, 'getAttribute'):
                rev = node.getAttribute('revision')
                if rev is not None:
                    return rev.encode('ascii')


    def _parseSVNEntries_8(self, entriesFile):
        """
        Given a readable file object which represents a .svn/entries file in
        format version 8, return the revision as a string.
        """
        entriesFile.readline()
        entriesFile.readline()
        entriesFile.readline()
        return entriesFile.readline().strip()
    
    
    # Add handlers for version 9 and 10 formats, which are the same as
    # version 8 as far as revision information is concerned.
    _parseSVNEntries_9 = _parseSVNEntries_8
    _parseSVNEntriesTenPlus = _parseSVNEntries_8


    def _getSVNVersion(self):
        """
        Figure out the SVN revision number based on the existance of
        <package>/.svn/entries, and its contents. This requires discovering the
        format version from the 'format' file and parsing the entries file
        accordingly.

        @return: None or string containing SVN Revision number.
        """
        mod = sys.modules.get(self.package)
        if mod:
            svn = os.path.join(os.path.dirname(mod.__file__), '.svn')
            if not os.path.exists(svn):
                # It's not an svn working copy
                return None

            formatFile = os.path.join(svn, 'format')
            if os.path.exists(formatFile):
                # It looks like a less-than-version-10 working copy.
                format = file(formatFile).read().strip()
                parser = getattr(self, '_parseSVNEntries_' + format, None)
            else:
                # It looks like a version-10-or-greater working copy, which
                # has version information in the entries file.
                parser = self._parseSVNEntriesTenPlus

            if parser is None:
                return 'Unknown'

            entriesFile = os.path.join(svn, 'entries')
            entries = file(entriesFile)
            try:
                try:
                    return parser(entries)
                finally:
                    entries.close()
            except:
                return 'Unknown'


    def _formatSVNVersion(self):
        ver = self._getSVNVersion()
        if ver is None:
            return ''
        return ' (SVN r%s)' % (ver,)



def getVersionString(version):
    """
    Get a friendly string for the given version object.

    @param version: A L{Version} object.
    @return: A string containing the package and short version number.
    """
    result = '%s %s' % (version.package, version.short())
    return result



import unittest

import StringIO

from amfast.context import (ContextError, Idx, Ref,
    DecoderContext, EncoderContext)

class ContextTestCase(unittest.TestCase):
    def setUp(self):
        self.test_string = 'tester'

    def testIdxMap(self):
        count = 64

        idx = Idx()
        for i in xrange(0, count):
            idx.map('%s' % i)

        for i in xrange(0, count):
            self.assertEquals('%s' % i, idx.ret(i))

    def testBadIndexRaisesException(self):
        idx = Idx()
        idx.map(self.test_string)
        self.assertRaises(ContextError, idx.ret, 1)

    def testRefMap(self):
        count = 64

        ref = Ref()
        strs = [str(i) for i in range(0, count)]
        for i, val in enumerate(strs):
            self.assertEquals(i, ref.map(val))

        for i, val in enumerate(strs):
            self.assertEquals(i, ref.ret(val))

    def testUnmappedObjReturnsNegative(self):
        ref = Ref()
        self.assertEquals(-1, ref.ret(self.test_string))

    def _testDecoderContext(self, con):
        self.assertEquals("Buffer", con.buffer.__class__.__name__)
        self.assertEquals("ClassDefMapper", con.class_def_mapper.__class__.__name__)
        self.assertEquals("Idx", con.obj_refs.__class__.__name__)
        self.assertEquals("Idx", con.string_refs.__class__.__name__)
        self.assertEquals("Idx", con.class_refs.__class__.__name__)

    def _testRead(self, con):
        self.assertEquals(self.test_string[0], con.read(1))
        self.assertEquals(self.test_string[1:len(self.test_string)], con.read(len(self.test_string) - 1 ))

    def testDecoderContext(self):
        self._testDecoderContext(DecoderContext(self.test_string, amf3=True))

    def testCopyDecoderContext(self):
        con = DecoderContext(self.test_string, amf3=True)
        con_2 = con.copy(amf3=True)
        self._testDecoderContext(con_2)

    def testReadString(self):
        self._testRead(DecoderContext(self.test_string))

    def testReadStream(self):
        stream = StringIO.StringIO(self.test_string)
        self._testRead(DecoderContext(stream))

    def _testEncoderContext(self, con):
        self.assertEquals("Buffer", con.buffer.__class__.__name__)
        self.assertEquals("ClassDefMapper", con.class_def_mapper.__class__.__name__)
        self.assertEquals("Ref", con.obj_refs.__class__.__name__)
        self.assertEquals("Ref", con.string_refs.__class__.__name__)
        self.assertEquals("Ref", con.class_refs.__class__.__name__)

    def _testWrite(self, con):
        test = self.test_string[0]
        con.write(test)
        self.assertEquals(test, con.buffer.getvalue())

        test = self.test_string[1:len(self.test_string)]
        con.write(test)
        self.assertEquals(self.test_string, con.buffer.getvalue())

    def testEncoderContext(self):
         self._testEncoderContext(EncoderContext(amf3=True))

    def testWriteString(self):
        self._testWrite(EncoderContext())

    def testWriteStream(self):
        self._testWrite(EncoderContext(StringIO.StringIO()))

    def testCopyEncoderContext(self):
        con = EncoderContext(amf3=True)
        con_2 = con.copy()
        self._testEncoderContext(con_2)

def suite():
    return unittest.TestLoader().loadTestsFromTestCase(ContextTestCase)

if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())



import time
import cPickle as pickle

import sqlalchemy as sa
from sqlalchemy.sql import func, and_

if sa.__version__.startswith('0.5'):
    BINARY_TYPE = sa.Binary
elif sa.__version__.startswith('0.6'):
    BINARY_TYPE = sa.LargeBinary

from connection import Connection
from connection_manager import NotConnectedError, ConnectionManager, SessionAttrError

class SaConnectionManager(ConnectionManager):
    """Manages connections in a database, uses SqlAlchemy to talk to the DB."""

    def __init__(self, engine, metadata, connection_class=Connection, connection_params=None,
                 table_prefix=''):
        ConnectionManager.__init__(self, connection_class=connection_class,
            connection_params=connection_params)

        self.engine = engine
        self.metadata = metadata
        self.table_prefix = table_prefix and "%s_" % table_prefix.rstrip('_') or table_prefix
        self.mapTables()

    def reset(self):
        db = self.getDb()
        db.execute(self.session_attrs.delete())
        db.execute(self.connections.delete())
        db.close()

    def mapTables(self):
        self.connections = sa.Table('%sconnections' % self.table_prefix, self.metadata,
            sa.Column('id', sa.String(36), primary_key=True),
            sa.Column('channel_name', sa.String(128), nullable=False),
            sa.Column('timeout', sa.Float(), nullable=False),
            sa.Column('connected', sa.Boolean(), nullable=False),
            sa.Column('last_active', sa.Float(), nullable=False),
            sa.Column('last_polled', sa.Float(), nullable=False, default=0.0),
            sa.Column('authenticated', sa.Boolean(), nullable=False, default=False),
            sa.Column('flex_user', sa.String(128), nullable=True),
            sa.Column('notify_func_id', sa.Integer(), nullable=True)
        )

        self.session_attrs = sa.Table('%ssession_attrs' % self.table_prefix, self.metadata,
            sa.Column('connection_id', sa.String(36),
                sa.ForeignKey('%sconnections.id' % self.table_prefix),
                primary_key=True, index=True),
            sa.Column('name', sa.String(128), primary_key=True),
            sa.Column('value', BINARY_TYPE(), nullable=False)
        )

    def createTables(self):
       db = self.getDb()
       self.connections.create(db, checkfirst=True)
       self.session_attrs.create(db, checkfirst=True)
       db.close()

    def getDb(self):
        return self.engine.connect()

    def loadConnection(self, connection_id):
        s = sa.select([self.connections.c.channel_name, self.connections.c.timeout],
            self.connections.c.id==connection_id)
        db = self.getDb()
        result = db.execute(s)
        row = result.fetchone()
        db.close()

        if row is None:
            raise NotConnectedError("Connection '%s' is not connected." % connection_id)
        
        return self.connection_class(self, row[self.connections.c.channel_name],
            connection_id, row[self.connections.c.timeout])

    def initConnection(self, connection, channel):
        ins = self.connections.insert().values(
            id=connection.id,
            channel_name=connection.channel_name,
            timeout=connection.timeout,
            connected=True,
            last_active=time.time() * 1000,
            last_polled=0.0,
            authenticated=False,
        )

        db = self.getDb()
        db.execute(ins)
        db.close()

    def getConnectionCount(self, channel_name):
        s = sa.select([sa.sql.func.count(self.connections.c.id)])
        db = self.getDb()
        result = db.execute(s)
        row = result.fetchone()
        db.close()

        return row[0]

    def iterConnectionIds(self):
        s = sa.select([self.connections.c.id])
        db = self.getDb()
        result = db.execute(s)
        for row in result:
            yield row[self.connections.c.id]
        db.close()

    # --- proxies for connection properties --- #

    def getConnected(self, connection):
        s = sa.select([self.connections.c.connected], self.connections.c.id==connection.id)
        db = self.getDb()
        result = db.execute(s)
        row =result.fetchone()
        db.close()

        if row is None:
            return False

        return row[self.connections.c.connected]

    def getLastActive(self, connection):
        s = sa.select([self.connections.c.last_active], self.connections.c.id==connection.id)
        db = self.getDb()
        result = db.execute(s)
        row = result.fetchone()
        db.close()

        return row[self.connections.c.last_active]

    def getLastPolled(self, connection):
        s = sa.select([self.connections.c.last_polled], self.connections.c.id==connection.id)
        db = self.getDb()
        result = db.execute(s)
        row = result.fetchone()
        db.close()

        return row[self.connections.c.last_polled]

    def getAuthenticated(self, connection):
        s = sa.select([self.connections.c.authenticated], self.connections.c.id==connection.id)
        db = self.getDb()
        result = db.execute(s)
        row = result.fetchone()
        db.close()

        return row[self.connections.c.authenticated]

    def getFlexUser(self, connection):
        s = sa.select([self.connections.c.flex_user], self.connections.c.id==connection.id)
        db = self.getDb()
        result = db.execute(s)
        row = result.fetchone()
        db.close()

        return row[self.connections.c.flex_user]

    def getNotifyFunc(self, connection):
        s = sa.select([self.connections.c.notify_func_id], self.connections.c.id==connection.id)
        db = self.getDb()
        result = db.execute(s)
        row = result.fetchone()
        db.close()

        if row is None:
            return None

        notify_func_id = notify_func_id = row[self.connections.c.notify_func_id]
        if notify_func_id is None:
            return None

        return connection._getNotifyFuncById(notify_func_id)

    # --- proxies for connection methods --- #

    def deleteConnection(self, connection):
        d = self.connections.delete().\
            where(self.connections.c.id==connection.id)
        db = self.getDb()
        db.execute(d)
        db.close()

        ConnectionManager.deleteConnection(self, connection)

    def connectConnection(self, connection):
        u = self.connections.update().\
            where(self.connections.c.id==connection.id).\
            values(connected=True)
        db = self.getDb()
        db.execute(u)
        db.close()

    def disconnectConnection(self, connection):
        u = self.connections.update().\
            where(self.connections.c.id==connection.id).\
            values(connected=False)
        db = self.getDb()
        db.execute(u)
        db.close()

    def touchConnection(self, connection):
        u = self.connections.update().\
            where(self.connections.c.id==connection.id).\
            values(last_active=time.time() * 1000)
        db = self.getDb()
        db.execute(u)
        db.close()

    def touchPolled(self, connection):
        u = self.connections.update().\
            where(self.connections.c.id==connection.id).\
            values(last_polled=time.time() * 1000)
        db = self.getDb()
        db.execute(u)
        db.close()

    def authenticateConnection(self, connection, user):
        u = self.connections.update().\
            where(self.connections.c.id==connection.id).\
            values(authenticated=True, flex_user=user)
        db = self.getDb()
        db.execute(u)
        db.close()

    def unAuthenticateConnection(self, connection):
        u = self.connections.update().\
            where(self.connections.c.id==connection.id).\
            values(authenticated=False, flex_user=None)
        db = self.getDb()
        db.execute(u)
        db.close()

    def setNotifyFunc(self, connection, func):
        u = self.connections.update().\
            where(self.connections.c.id==connection.id).\
            values(notify_func_id=connection._setNotifyFunc(func))
        db = self.getDb()
        db.execute(u)
        db.close()

    def unSetNotifyFunc(self, connection):
        s = sa.select([self.connections.c.notify_func_id], self.connections.c.id==connection.id)
        db = self.getDb()
        result = db.execute(s)
        row = result.fetchone()

        if row is None:
            return

        connection._delNotifyFunc(row[self.connections.c.notify_func_id])

        u = self.connections.update().\
            where(self.connections.c.id==connection.id).\
            values(notify_func_id=None)
        db.execute(u)
        db.close()

    def getConnectionSessionAttr(self, connection, name):
        s = sa.select([self.session_attrs.c.value],
            and_(self.session_attrs.c.connection_id==connection.id,
                self.session_attrs.c.name==name))
        db = self.getDb()
        result = db.execute(s)
        row = result.fetchone()
        db.close()

        if row is None:
            raise SessionAttrError("Attribute '%s' not found." % name)

        return pickle.loads(str(row[self.session_attrs.c.value]))

    def setConnectionSessionAttr(self, connection, name, val):
        s = sa.select([self.session_attrs.c.connection_id],
            and_(self.session_attrs.c.connection_id==connection.id,
                self.session_attrs.c.name==name))
        db = self.getDb()
        result = db.execute(s)
        row = result.fetchone()
        
        if row is None: 
            statement = self.session_attrs.insert().values(
                connection_id=connection.id,
                name=name, value=pickle.dumps(val))
        else:
            statement = self.connections.update().\
                where(and_(self.session_attrs.c.connection_id==connection.id,
                    self.session_attrs.c.name==name)).\
                values(value=pickle.dumps(val))

        db.execute(statement)
        db.close()

    def delConnectionSessionAttr(self, connection, name):
        d = self.session_attrs.delete().\
            where(and_(self.session_attrs.c.connection_id==connection.id,
                self.session_attrs.c.name==name))

        db = self.getDb()
        db.execute(d)
        db.close()



# -*- coding: utf-8 -*-
import unittest

from amfast.context import EncoderContext
import amfast.encode as encode
import amfast.class_def as class_def

class Amf3EncoderTestCase(unittest.TestCase):
    class Spam(object):
        def __init__(self):
            self.spam = 'eggs'

    def setUp(self):
        self.class_mapper = class_def.ClassDefMapper()

    def testNull(self):
        buf = encode.encode(None, EncoderContext(amf3=True))
        self.assertEquals('\x01', buf)

    def testFalse(self):
        buf = encode.encode(False, EncoderContext(amf3=True))
        self.assertEquals('\x02', buf)

    def testTrue(self):
        buf = encode.encode(True, EncoderContext(amf3=True))
        self.assertEquals('\x03', buf)

    def testInt(self):
        tests = {
            0: '\x04\x00',
            0x35: '\x04\x35',
            0x7f: '\x04\x7f',
            0x80: '\x04\x81\x00',
            0xd4: '\x04\x81\x54',
            0x3fff: '\x04\xff\x7f',
            0x4000: '\x04\x81\x80\x00',
            0x1a53f: '\x04\x86\xca\x3f',
            0x1fffff: '\x04\xff\xff\x7f',
            0x200000: '\x04\x80\xc0\x80\x00',
            -0x01: '\x04\xff\xff\xff\xff',
            -0x2a: '\x04\xff\xff\xff\xd6',
            0xfffffff: '\x04\xbf\xff\xff\xff'
        }

        for integer, encoding in tests.iteritems():
            buf = encode.encode(integer, EncoderContext(amf3=True))
            self.assertEquals(encoding, buf)

    def testFloat(self):
        tests = {
            0.1: '\x05\x3f\xb9\x99\x99\x99\x99\x99\x9a',
            0.123456789: '\x05\x3f\xbf\x9a\xdd\x37\x39\x63\x5f'
        }

        for number, encoding in tests.iteritems():
            buf = encode.encode(number, EncoderContext(amf3=True))
            self.assertEquals(encoding, buf)

    def testLong(self):
        tests = {
            0x10000000: '\x05\x41\xb0\x00\x00\x00\x00\x00\x00',
            -0x10000001: '\x05\xc1\xb0\x00\x00\x01\x00\x00\x00',
            -0x10000000: '\x04\xc0\x80\x80\x00'
        }

        for number, encoding in tests.iteritems():
            buf = encode.encode(number, EncoderContext(amf3=True))
            self.assertEquals(encoding, buf)

    def testUnicode(self):
        tests = {
            u'': '\x06\x01',
            u'hello': '\x06\x0bhello',
            u'ᚠᛇᚻ': '\x06\x13\xe1\x9a\xa0\xe1\x9b\x87\xe1\x9a\xbb'
        }

        for string, encoding in tests.iteritems():
            buf = encode.encode(string, EncoderContext(amf3=True))
            self.assertEquals(encoding, buf)

    def testUnicodeRefs(self):
        hello = u'hello'
        test = [hello, hello, '', hello]

        result = '\x09\x09\x01' # array header
        result += '\x06\x0bhello' # array element 1 (hello encoded)
        result += '\x06\x00' #array element 2 (reference to hello)
        result += '\x06\x01' #array element 3 (empty string)
        result += '\x06\x00' #array element 4 (reference to hello)

        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals(result, buf)

    def testRefsOff(self):
        hello = u'hello'
        test = [hello, hello, '', hello]

        result = '\x09\x09\x01' # array header
        result += '\x06\x0bhello' # array element 1 (hello encoded)
        result += '\x06\x0bhello' #array element 2 (hello encoded)
        result += '\x06\x01' #array element 3 (empty string)
        result += '\x06\x0bhello' #array element 4 (hello encoded)

        buf = encode.encode(test, EncoderContext(use_references=False, amf3=True))
        self.assertEquals(result, buf)

    def testString(self):
        tests = {
            '': '\x06\x01',
            'hello': '\x06\x0bhello'
        }

        for string, encoding in tests.iteritems():
            buf = encode.encode(string, EncoderContext(amf3=True))
            self.assertEquals(encoding, buf)

    def testStringRefs(self):
        hello = 'hello'
        test = [hello, hello, '', hello]

        result = '\x09\x09\x01' # array header
        result += '\x06\x0bhello' # array element 1 (hello encoded)
        result += '\x06\x00' #array element 2 (reference to hello)
        result += '\x06\x01' #array element 3 (empty string)
        result += '\x06\x00' #array element 4 (reference to hello)

        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals(result, buf)

    def testTuple(self):
        test = (0, 1, 2, 3);

        result = '\x09\x09\x01' #array header
        result += '\x04\x00' #array element 1
        result += '\x04\x01' #array element 2
        result += '\x04\x02' #array element 3
        result += '\x04\x03' #array element 4

        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals(result, buf);

    def testTupleForceNoProxy(self):
        from amfast.class_def.as_types import AsNoProxy
        test = (0, 1, 2, 3);

        result = '\x09\x09\x01' #array header
        result += '\x04\x00' #array element 1
        result += '\x04\x01' #array element 2
        result += '\x04\x02' #array element 3
        result += '\x04\x03' #array element 4

        buf = encode.encode(AsNoProxy(test), EncoderContext(amf3=True,
            use_collections=True))
        self.assertEquals(result, buf);

    def testTupleRefs(self):
        test_tuple = (0, 1, 2, 3);
        test = [test_tuple, test_tuple]

        result = '\x09\x05\x01' # array header
        result += '\x09\x09\x01\x04\x00\x04\x01\x04\x02\x04\x03' # array element 1 (test_tuple encoded)
        result += '\x09\x02' # array element 2 (reference to test_tuple)

        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals(result, buf)

    def testTupleAsCollection(self):
        test = (0, 1, 2, 3);

        result = '\x0A\x07\x43flex.messaging.io.ArrayCollection' # Object header 
        result += '\x09\x09\x01' #array header
        result += '\x04\x00' #array element 1
        result += '\x04\x01' #array element 2
        result += '\x04\x02' #array element 3
        result += '\x04\x03' #array element 4

        buf = encode.encode(test, EncoderContext(use_collections=True, amf3=True))
        self.assertEquals(result, buf)

    def testTupleAsForcedCollection(self):
        from amfast.class_def.as_types import AsProxy
        test = (0, 1, 2, 3);

        result = '\x0A\x07\x43flex.messaging.io.ArrayCollection' # Object header 
        result += '\x09\x09\x01' #array header
        result += '\x04\x00' #array element 1
        result += '\x04\x01' #array element 2
        result += '\x04\x02' #array element 3
        result += '\x04\x03' #array element 4

        buf = encode.encode(AsProxy(test), EncoderContext(use_collections=False, amf3=True))
        self.assertEquals(result, buf)


    def testArrayCollectionRef(self):
        test_tuple = (0, 1, 2, 3);
        test = [test_tuple, test_tuple]

        result = '\x0A\x07\x43flex.messaging.io.ArrayCollection\x09\x05\x01' # array header
        result += '\x0A\x01\x09\x09\x01\x04\x00\x04\x01\x04\x02\x04\x03' # array element 1 (test_tuple encoded)
        result += '\x0A\x04' # array element 2 (reference to test_tuple)

        buf = encode.encode(test, EncoderContext(use_collections=True, amf3=True))
        self.assertEquals(result, buf)

    def testList(self):
        test = [0, 1, 2, 3];

        result = '\x09\x09\x01' #array header
        result += '\x04\x00' #array element 1
        result += '\x04\x01' #array element 2
        result += '\x04\x02' #array element 3
        result += '\x04\x03' #array element 

        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals(result, buf);

    def testListRefs(self):
        test_list = [0, 1, 2, 3];
        test = (test_list, test_list)

        result = '\x09\x05\x01' # array header
        result += '\x09\x09\x01\x04\x00\x04\x01\x04\x02\x04\x03' # array element 1 (test_list encoded)
        result += '\x09\x02' # array element 2 (reference to test_list)

        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals(result, buf)

    def testListAsCollection(self):
        test = [0, 1, 2, 3];

        result = '\x0A\x07\x43flex.messaging.io.ArrayCollection' # Object header 
        result += '\x09\x09\x01' #array header
        result += '\x04\x00' #array element 1
        result += '\x04\x01' #array element 2
        result += '\x04\x02' #array element 3
        result += '\x04\x03' #array element 4

        buf = encode.encode(test, EncoderContext(use_collections=True, amf3=True))
        self.assertEquals(result, buf)

    def testDict(self):
        result = '\x0A\x0B\x01' # Object header
        result += '\x09spam' # key
        result += '\x06\x09eggs' #value
        result += '\x01' # empty string terminator

        buf = encode.encode({'spam': 'eggs'}, EncoderContext(amf3=True))
        self.assertEquals(result, buf)

    def testDictForceNoProxy(self):
        from amfast.class_def.as_types import AsNoProxy

        result = '\x0A\x0B\x01' # Object header
        result += '\x09spam' # key
        result += '\x06\x09eggs' #value
        result += '\x01' # empty string terminator

        buf = encode.encode(AsNoProxy({'spam': 'eggs'}), EncoderContext(amf3=True,
            use_proxies=True))
        self.assertEquals(result, buf)

    def testDictRef(self):
        test_dict = {'spam': 'eggs'};
        test = (test_dict, test_dict)

        result = '\x09\x05\x01' #array header
        result += '\x0A\x0B\x01\x09spam\x06\x09eggs\x01' #array element 1 (test_dict encoded)
        result += '\x0A\x02' # array element 2 (reference to test_dict)

        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals(result, buf)

    def testDictAsObjectProxy(self):
        result = '\x0A\x07\x3Bflex.messaging.io.ObjectProxy' # Object header 
        result += '\x0A\x0B\x01' # Object header
        result += '\x09spam' # key
        result += '\x06\x09eggs' #value
        result += '\x01' # empty string terminator

        buf = encode.encode({'spam': 'eggs'}, EncoderContext(use_proxies=True, amf3=True))
        self.assertEquals(result, buf)

    def testDictForcedObjectProxy(self):
        from amfast.class_def.as_types import AsProxy

        result = '\x0A\x07\x3Bflex.messaging.io.ObjectProxy' # Object header 
        result += '\x0A\x0B\x01' # Object header
        result += '\x09spam' # key
        result += '\x06\x09eggs' #value
        result += '\x01' # empty string terminator

        buf = encode.encode(AsProxy({'spam': 'eggs'}), EncoderContext(use_proxies=False, amf3=True))
        self.assertEquals(result, buf)

    def testObjectProxyRef(self):
        test_dict = {'spam': 'eggs'}
        test = (test_dict, test_dict)

        result = '\x09\x05\x01' #array header
        result += '\x0A\x07\x3Bflex.messaging.io.ObjectProxy\x0A\x0B\x01\x09spam\x06\x09eggs\x01' #array element 1 (test_dict encoded)
        result += '\x0A\x02' # array element 2 (reference to test_dict)

        buf = encode.encode(test, EncoderContext(use_proxies=True, amf3=True))
        self.assertEquals(result, buf)

    def testDate(self):
        import datetime

        test = datetime.datetime(2005, 3, 18, 1, 58, 31)
        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals('\x08\x01Bp+6!\x15\x80\x00', buf)

        test = datetime.date(2003, 12, 1)
        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals('\x08\x01Bo%\xe2\xb2\x80\x00\x00', buf)

    def testDateReferences(self):
        import datetime

        test_date = datetime.datetime(2005, 3, 18, 1, 58, 31)
        test = [test_date, test_date]

        result = '\x09\x05\x01' #array header
        result += '\x08\x01Bp+6!\x15\x80\x00' #array element 1 (test_date encoded)
        result += '\x08\x02' # array element 2 (reference to test_date)

        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals(result, buf)

    def testXml(self):
        import xml.dom.minidom

        document = """
           <test>
            <test_me>tester</test_me>
           </test>
        """

        result = '\x0B' # XML header
        result += '\x81\x2B' # String header
        result += '<?xml version="1.0" ?><test>\n            <test_me>tester</test_me>\n           </test>' # encoded XML

        dom = xml.dom.minidom.parseString(document)
        buf = encode.encode(dom, EncoderContext(amf3=True))
        self.assertEquals(result, buf)

    def testXmlRef(self):
        import xml.dom.minidom

        document = """
           <test>
            <test_me>tester</test_me>
           </test>
        """

        dom = xml.dom.minidom.parseString(document)
        test = [dom, dom]

        result = '\x09\x05\x01' #array header
        result += '\x0B\x81\x2B<?xml version="1.0" ?><test>\n            <test_me>tester</test_me>\n           </test>' # array element 1 (encoded dom)
        result += '\x0B\x02' # array element 2 (reference to dom)

        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals(result, buf)

    def testLegacyXml(self):
        import xml.dom.minidom

        document = """
           <test>
            <test_me>tester</test_me>
           </test>
        """

        result = '\x07' # XML header
        result += '\x81\x2B' # String header
        result += '<?xml version="1.0" ?><test>\n            <test_me>tester</test_me>\n           </test>' # encoded XML

        dom = xml.dom.minidom.parseString(document)
        buf = encode.encode(dom, EncoderContext(use_legacy_xml=True, amf3=True))
        self.assertEquals(result, buf)

    def testAnonObj(self):
        test = self.Spam()

        result = '\x0A\x0B\x01' # Object header
        result += '\x09spam' # key
        result += '\x06\x09eggs' #value
        result += '\x01' # empty string terminator

        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals(result, buf)

    def testAnonObjRef(self):
        test_obj = self.Spam()
        test = (test_obj, test_obj)

        result = '\x09\x05\x01' #array header
        result += '\x0A\x0B\x01\x09spam\x06\x09eggs\x01' #array element 1 (test_obj encoded)
        result += '\x0A\x02' # array element 2 (reference to test_obj)

        buf = encode.encode(test, EncoderContext(amf3=True))
        self.assertEquals(result, buf)

    def testStaticObj(self):
        self.class_mapper.mapClass(class_def.ClassDef(self.Spam, 'alias.spam', ('spam',)))
        test = self.Spam()

        result = '\x0A\x13\x15alias.spam'
        result += '\x09spam' # static attr definition
        result += '\x06\x09eggs' # static attrs

        buf = encode.encode(test, EncoderContext(\
            class_def_mapper=self.class_mapper, amf3=True))
        self.class_mapper.unmapClass(self.Spam)

        self.assertEquals(result, buf)

    def testClassRef(self):
        self.class_mapper.mapClass(class_def.ClassDef(self.Spam, 'alias.spam', ('spam',)))
        test_obj_1 = self.Spam()
        test_obj_2 = self.Spam()
        test_obj_2.spam = 'foo'
        test = (test_obj_1, test_obj_2)

        result = '\x09\x05\x01' #array header
        result += '\x0A\x13\x15alias.spam\x09spam\x06\x09eggs' # array element 1
        result += '\x0A\x01\x06\x07foo' # array element 2

        buf = encode.encode(test, EncoderContext(\
            class_def_mapper=self.class_mapper, amf3=True))
        self.class_mapper.unmapClass(self.Spam)

        self.assertEquals(result, buf)

    def testTypedObjectRef(self):
        self.class_mapper.mapClass(class_def.ClassDef(self.Spam, 'alias.spam', ('spam',)))
        test_obj = self.Spam()
        test = (test_obj, test_obj)

        result = '\x09\x05\x01' #array header
        result += '\x0A\x13\x15alias.spam\x09spam\x06\x09eggs' # test_obj_encoded
        result += '\x0A\x02' # array element 2 (reference to test_obj)

        buf = encode.encode(test, EncoderContext(\
            class_def_mapper=self.class_mapper, amf3=True))
        self.class_mapper.unmapClass(self.Spam)

        self.assertEquals(result, buf)

    def testDynamicObj(self):
        self.class_mapper.mapClass(class_def.DynamicClassDef(self.Spam, 'alias.spam', ()))
        test = self.Spam()
        
        result = '\x0A\x0B\x15alias.spam'
        result += '\x09spam\x06\x09eggs\x01' # dynamic attrs

        buf = encode.encode(test, EncoderContext(\
            class_def_mapper=self.class_mapper, amf3=True))
        self.class_mapper.unmapClass(self.Spam)

        self.assertEquals(result, buf)

    def testStaticDynamicObj(self):
        self.class_mapper.mapClass(class_def.DynamicClassDef(self.Spam, 'alias.spam', ('spam',)))
        test = self.Spam()
        test.ham = 'foo'

        result = '\x0A\x1B\x15alias.spam' # obj header
        result += '\x09spam' # static attr definition
        result += '\x06\x09eggs' # static attrs
        result += '\x07ham\x06\x07foo\x01' #dynamic attrs

        buf = encode.encode(test, EncoderContext(\
            class_def_mapper=self.class_mapper, amf3=True))
        self.class_mapper.unmapClass(self.Spam)

        self.assertEquals(result, buf)

    def testExternalizable(self):
        custom_encoding = '\x01\x02\x03\x04\x05'

        class ExternClass(class_def.ExternClassDef):
            def writeExternal(self, obj, context):
                context.buffer.write(custom_encoding)

        self.class_mapper.mapClass(ExternClass(self.Spam, 'alias.spam', ('spam',)))
        test = self.Spam()

        result = '\x0A\x07\x15alias.spam'
        result += custom_encoding # raw bytes

        buf = encode.encode(test, EncoderContext(\
            class_def_mapper=self.class_mapper, amf3=True))
        self.class_mapper.unmapClass(self.Spam)

        self.assertEquals(result, buf)

    def testUnknownRaisesError(self):
        self.assertRaises(AttributeError, encode.encode, set())

def suite():
    return unittest.TestLoader().loadTestsFromTestCase(Amf3EncoderTestCase)

if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())



import threading

class WorkerTask(object):
    """A task to be performed by the ThreadPool."""

    def __init__(self, function, args=(), kwargs={}):
        self.function = function
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        self.function(*self.args, **self.kwargs)

class RepeatingThread(threading.Thread):
    """A thread that repeats the same task over and over."""

    def __init__(self, duration=None, task=None):
        threading.Thread.__init__(self)
        self.setDaemon(True)
        self.duration = duration
        self.task = task
        self._stop = False
        self._event = None

    def run(self):
        if self._stop is False:
            self._event = threading.Event()
            while True:
                self._event.wait(self.duration)
                if self._stop is True:
                    return
                self.task()

    def stop(self):
        self._stop = True
        if self._event is not None:
            self._event.set()

class WorkerThread(threading.Thread):
    """A thread managed by a thread pool."""

    def __init__(self, pool):
        threading.Thread.__init__(self)
        self.setDaemon(True)
        self.pool = pool
        self.busy = False
        self._started = False
        self._event = None

    def work(self):
        if self._started is True:
            if self._event is not None and not self._event.isSet():
                self._event.set()
        else:
            self._started = True
            self.start()

    def run(self):
        while True:
            self.busy = True
            while len(self.pool._tasks) > 0:
                try:
                    task = self.pool._tasks.pop()
                    task()
                except IndexError:
                    # Just in case another thread grabbed the task 1st.
                    pass

            # Sleep until needed again
            self.busy = False
            if self._event is None:
                self._event = threading.Event()
            else:
                self._event.clear()
            self._event.wait()

class ThreadPool(object):
    """Executes queued tasks in the background."""

    def __init__(self, max_pool_size=10):
        self.max_pool_size = max_pool_size
        self._threads = []
        self._tasks = [] 

    def _addTask(self, task):
        self._tasks.append(task)

        worker_thread = None
        for thread in self._threads:
            if thread.busy is False:
                worker_thread = thread
                break

        if worker_thread is None and len(self._threads) <= self.max_pool_size:
            worker_thread = WorkerThread(self)
            self._threads.append(worker_thread)

        if worker_thread is not None:
            worker_thread.work()

    def addTask(self, function, args=(), kwargs={}):
        self._addTask(WorkerTask(function, args, kwargs))

class GlobalThreadPool(object):
    """ThreadPool Singleton class."""

    _instance = None

    def __init__(self):
        """Create singleton instance """

        if GlobalThreadPool._instance is None:
            # Create and remember instance
            GlobalThreadPool._instance = ThreadPool()

    def __getattr__(self, attr):
        """ Delegate get access to implementation """
        return getattr(self._instance, attr)

    def __setattr__(self, attr, val):
        """ Delegate set access to implementation """
        return setattr(self._instance, attr, val)



import unittest
import logging
import sys

import amfast
from amfast import remoting, logger
from amfast.encoder import Encoder
from amfast.decoder import Decoder
from amfast.remoting import ServiceMapper, flex_messages as messaging
from amfast.remoting.channel import ChannelSet, Channel

#handler = logging.StreamHandler(sys.stdout)
#handler.setLevel(logging.DEBUG)
#logger.addHandler(handler)
#amfast.log_debug = True

class _mappingObj(object):
    """Used to test Target.mapObject."""
    def __init__(self):
        self.foo = 'bar'

    def _private(self):
        pass

    def public(self):
        pass

class RemotingTestCase(unittest.TestCase):

    def setUp(self):
        self.service_mapper = ServiceMapper()

        # Map example function to target
        self.target_name = 'Credentials'
        target = remoting.CallableTarget(self.login, self.target_name)

        # Map header targets
        self.service_mapper.packet_header_service.mapTarget(target)

        # Map message targets
        self.service_name = 'login'
        service = remoting.Service(self.service_name)
        service.mapTarget(target)
        self.service_mapper.mapService(service)

        # Valid values for the target
        self.arg = {'userid': 'userid', 'password': 'password'}

        # Setup Encoder and Decoder
        self.channel_set = ChannelSet(service_mapper=self.service_mapper)
        self.channel = Channel('amf_rpc')
        self.channel_set.mapChannel(self.channel)

    def tearDown(self):
        pass

    def login(self, credentials):
        self.assertEquals(self.arg['userid'], credentials['userid'])
        self.assertEquals(self.arg['password'], credentials['password'])
        return True

    def testMapObjectExcludesAttr(self):
        test_service = remoting.Service('test')
        remoting.CallableTarget.mapObject(test_service, _mappingObj())
        self.assertEquals(None, test_service.getTarget('foo'))

    def testMapObjectExcludesPrivate(self):
        test_service = remoting.Service('test')
        remoting.CallableTarget.mapObject(test_service, _mappingObj())
        self.assertEquals(None, test_service.getTarget('_private'))

    def testMapObjectExcludesSpecial(self):
        test_service = remoting.Service('test')
        remoting.CallableTarget.mapObject(test_service, _mappingObj())
        self.assertEquals(None, test_service.getTarget('__init__'))

    def testMapObjectIncludesMethods(self):
        test_service = remoting.Service('test')
        remoting.CallableTarget.mapObject(test_service, _mappingObj(), map_private=True)
        self.assertTrue(isinstance(test_service.getTarget('_private'), remoting.Target))
        self.assertTrue(isinstance(test_service.getTarget('public'), remoting.Target))

    def testDecodeRpcPacket(self):
        encoded = '\x00\x00' # AMF0 version marker
        encoded += '\x00\x02' # Header count (2)
        encoded += '\x00\x04spam\x00\x00\x00\x00\x07\x02\x00\x04eggs' # Optional Header 
        encoded += '\x00\x04eggs\x01\x00\x00\x00\x07\x02\x00\x04spam' # Required Header
        encoded += '\x00\x02' # Body count (2)
        #1st message body
        encoded += '\x00\x04spam' # Body target
        encoded += '\x00\x04eggs' # Body response
        encoded += '\x00\x00\x00\x07' # Body byte length
        encoded += '\x0A\x00\x00\x00\x01' # Body start - 1 element array header
        encoded += '\x01\x01' # Body[0] 1 value
        #2nd message body
        encoded += '\x00\x04eggs' # Body 2 target
        encoded += '\x00\x04spam' # Body 2 response
        encoded += '\x00\x00\x00\x07' # Body 2 byte length
        encoded += '\x0A\x00\x00\x00\x01' # Body start - 1 element array header
        encoded += '\x01\x00' # Body 2 value

        packet = self.channel.endpoint.decodePacket(encoded)

        # version
        self.assertEquals(packet.FLASH_8, packet.client_type)

        # headers
        self.assertEquals('eggs', packet.headers[0].value)
        self.assertEquals(False, packet.headers[0].required)
        self.assertEquals('spam', packet.headers[0].name)

        self.assertEquals('spam', packet.headers[1].value)
        self.assertEquals(True, packet.headers[1].required)
        self.assertEquals('eggs', packet.headers[1].name)

        # messages
        self.assertEquals('spam', packet.messages[0].target)
        self.assertEquals('eggs', packet.messages[0].response)
        self.assertEquals(True, packet.messages[0].body[0])

        self.assertEquals('eggs', packet.messages[1].target)
        self.assertEquals('spam', packet.messages[1].response)
        self.assertEquals(False, packet.messages[1].body[0])

    def testDecodeReponsePacket(self):
        encoded = '\x00\x00' # AMF0 version marker
        encoded += '\x00\x01' # Header count
        encoded += '\x00\x04spam\x00\x00\x00\x00\x07\x02\x00\x04eggs' # Optional Header 
        encoded += '\x00\x01' # Body count
        #1st message body
        encoded += '\x00\x04spam' # Body target
        encoded += '\x00\x00' # Body response
        encoded += '\x00\x00\x00\x02' # Body byte length
        encoded += '\x01\x01' # Body value

        packet = self.channel.endpoint.decodePacket(encoded)

        # version
        self.assertEquals(packet.FLASH_8, packet.client_type)

        # headers
        self.assertEquals('eggs', packet.headers[0].value)
        self.assertEquals(False, packet.headers[0].required)
        self.assertEquals('spam', packet.headers[0].name)

        # messages
        self.assertEquals('spam', packet.messages[0].target)
        self.assertEquals('', packet.messages[0].response)
        self.assertEquals(True, packet.messages[0].body)

    def testRoundTripPacket(self):
        encoded = '\x00\x00' # AMF0 version marker
        encoded += '\x00\x02' # Header count (2)
        encoded += '\x00\x04spam\x00\x00\x00\x00\x07\x02\x00\x04eggs' # Optional Header 
        encoded += '\x00\x04eggs\x01\x00\x00\x00\x07\x02\x00\x04spam' # Required Header
        encoded += '\x00\x02' # Body count (2)
        #1st message body
        encoded += '\x00\x04spam' # Body target
        encoded += '\x00\x04eggs' # Body response
        encoded += '\x00\x00\x00\x07' # Body byte length
        encoded += '\x0A\x00\x00\x00\x01' # Body start - 1 element array header
        encoded += '\x01\x01' # Body[0] 1 value
        #2nd message body
        encoded += '\x00\x04eggs' # Body 2 target
        encoded += '\x00\x04spam' # Body 2 response
        encoded += '\x00\x00\x00\x07' # Body 2 byte length
        encoded += '\x0A\x00\x00\x00\x01' # Body start - 1 element array header
        encoded += '\x01\x00' # Body 2 value

        packet = self.channel.endpoint.decodePacket(encoded)
        encoded_packet = self.channel.endpoint.encodePacket(packet)
        self.assertEquals(encoded, encoded_packet)

    def testHeaderTarget(self):
        header = remoting.Header(self.target_name, False, self.arg)
        packet = remoting.Packet(headers=[header])
        packet.channel = self.channel
        packet.invoke()

    def testOldStyleTarget(self):
        qualified_name = self.service_name + '.' + self.target_name
        message = remoting.Message(target=qualified_name, response='/1',
            body=(self.arg,))
        packet = remoting.Packet(messages=[message])
        packet.channel = self.channel
        response = packet.invoke()
        
        self.assertEquals(1, len(response.messages))
        self.assertEquals(True, response.messages[0].target.endswith('onResult'))
        self.assertEquals('', response.messages[0].response)
        self.assertEquals(True, response.messages[0].body)

    def testOldStyleTargetFault(self):
        message = remoting.Message(target='bad_target', response='/1',
            body=(self.arg,))
        packet = remoting.Packet(messages=[message])
        packet.channel = self.channel
        response = packet.invoke()

        self.assertEquals(1, len(response.messages))
        self.assertEquals(True, response.messages[0].target.endswith('onStatus'))
        self.assertEquals('', response.messages[0].response)
        self.assertEquals(remoting.AsError, response.messages[0].body.__class__)

    def testRoTarget(self):
        outter_msg = remoting.Message(target='null', response='/1')
        inner_msg = messaging.RemotingMessage()
        inner_msg.destination = self.service_name
        inner_msg.operation = self.target_name
        inner_msg.headers = {self.target_name: self.arg, 'DSEndpoint': 'amf'}
        inner_msg.body = (self.arg,)
        inner_msg.messageId = '123'
        outter_msg.body = (inner_msg, )

        packet = remoting.Packet(messages=[outter_msg])
        packet.channel = self.channel
        response = packet.invoke()

        # Check outer msg
        self.assertEquals(1, len(response.messages))
        self.assertEquals(True, response.messages[0].target.endswith('onResult'))
        self.assertEquals('', response.messages[0].response)

        # Check inner msg
        self.assertEquals(True, response.messages[0].body.body)
        self.assertEquals('123', response.messages[0].body.correlationId)

    def testRoTargetFault(self):
        outter_msg = remoting.Message(target='null', response='/1')
        inner_msg = messaging.RemotingMessage()
        inner_msg.destination = 'fault'
        inner_msg.operation = self.target_name
        inner_msg.headers = {self.target_name: self.arg, 'DSEndpoint': 'amf'}
        inner_msg.body = (self.arg,)
        inner_msg.messageId = '123'
        outter_msg.body = (inner_msg, )

        packet = remoting.Packet(messages=[outter_msg])
        packet.channel = self.channel
        response = packet.invoke()

        # Check outer msg
        self.assertEquals(1, len(response.messages))
        self.assertEquals(True, response.messages[0].target.endswith('onStatus'))
        self.assertEquals('', response.messages[0].response)

        # Check inner msg
        self.assertEquals(messaging.ErrorMessage, response.messages[0].body.__class__)
        self.assertEquals('123', response.messages[0].body.correlationId)

    def testRoPing(self):
        outter_msg = remoting.Message(target='null', response='/1')
        inner_msg = messaging.CommandMessage()
        inner_msg.destination = self.service_name
        inner_msg.operation = messaging.CommandMessage.CLIENT_PING_OPERATION
        inner_msg.headers = {'DSEndpoint': 'amf'}
        inner_msg.body = ()
        inner_msg.messageId = '123'
        outter_msg.body = (inner_msg, )

        packet = remoting.Packet(messages=[outter_msg])
        packet.channel = self.channel
        response = packet.invoke()

        # Check outer msg
        self.assertEquals(1, len(response.messages))
        self.assertEquals(True, response.messages[0].target.endswith('onResult'))
        self.assertEquals('', response.messages[0].response)

        # Check inner msg
        self.assertEquals(messaging.AcknowledgeMessage, response.messages[0].body.__class__)
        self.assertEquals('123', response.messages[0].body.correlationId)

    def testProcessPacket(self):
        outter_msg = remoting.Message(target='null', response='/1')
        inner_msg = messaging.CommandMessage()
        inner_msg.destination = self.service_name
        inner_msg.operation = messaging.CommandMessage.CLIENT_PING_OPERATION
        inner_msg.headers = {'DSEndpoint': 'amf'}
        inner_msg.body = (None, )
        inner_msg.messageId = '123'
        outter_msg.body = (inner_msg, )

        packet = remoting.Packet(messages=[outter_msg])
        encoded_packet = self.channel.encode(packet)
        decoded_packet = self.channel.decode(encoded_packet)
        response = self.channel.invoke(decoded_packet)
        encoded_response = self.channel.encode(response)
        response = self.channel.endpoint.decodePacket(encoded_response)

        # Check outer msg
        self.assertEquals(1, len(response.messages))
        self.assertEquals(True, response.messages[0].target.endswith('onResult'))
        self.assertEquals('', response.messages[0].response)

        # Check inner msg
        self.assertEquals(messaging.AcknowledgeMessage, response.messages[0].body.__class__)
        self.assertEquals('123', response.messages[0].body.correlationId)

def suite():
    return unittest.TestLoader().loadTestsFromTestCase(RemotingTestCase)

if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())




import amfast
from gae_connection_manager import GaeConnectionManager
from gae_subscription_manager import GaeSubscriptionManager
from pyamf_endpoint import PyAmfEndpoint
from wsgi_channel import WsgiChannelSet, WsgiChannel

class GaeChannelSet(WsgiChannelSet):

    # Access this URL to clean-up stale connection data.
    CLEAN_URL = '_clean'

    def __init__(self, service_mapper=None, connection_manager=None,
        subscription_manager=None):

        if connection_manager is None:
            connection_manager = GaeConnectionManager()

        if subscription_manager is None:
            subscription_manager = GaeSubscriptionManager()

        WsgiChannelSet.__init__(self, service_mapper=service_mapper,
            connection_manager=connection_manager,
            subscription_manager=subscription_manager,
            clean_freq=100000)

    def __call__(self, environ, start_response):
        channel_name = environ['PATH_INFO'][1:]
        if channel_name == self.CLEAN_URL:
            self.clean()
            return

        channel = self.getChannel(channel_name)
        return channel(environ, start_response)

    def notifyConnections(self):
        pass

    def scheduleClean(self):
        amfast.logger.warn("Cron.yaml must be configured to access the URL: self.CLEAN_URL to clean-up any stale connections.")

class GaeChannel(WsgiChannel):
    def __init__(self, *args, **kwargs):
        # Overriden to set default endpoint as PyAmfEndpoint.
        if len(args) < 3 and ('endpoint' not in kwargs):
            kwargs['endpoint'] = PyAmfEndpoint()

        WsgiChannel.__init__(self, *args, **kwargs)



import pickle

from google.appengine.ext import db

import amfast
from subscription_manager import Subscription, SubscriptionManager

class GaeSubscription(db.Model, Subscription):
    """A client's subscription to a topic, persisted in a Google Datastore."""

    @classmethod
    def getKeyName(cls, connection_id, client_id, topic):
        return ':'.join((connection_id, client_id, topic))

    connection_id = db.StringProperty(required=True)
    client_id = db.StringProperty(required=True)
    topic = db.StringProperty(required=True)

class GaeMessageBody(db.Model):
    """Flex message body persisted in a Google Datastore."""

    p_message = db.BlobProperty(required=True)

class GaeMessageMetadata(db.Model):
    """Flex message metadata persisted in a Google Datastore."""

    time_to_live = db.FloatProperty(required=True)
    timestamp = db.FloatProperty(required=True)
    topic = db.StringProperty(required=True)
    message_body = db.ReferenceProperty(reference_class=GaeMessageBody, required=True)

class GaeSubscriptionManager(SubscriptionManager):
    """Stores subscriptions in Google DataStore."""

    def reset(self):
        query = GaeSubscription.all(keys_only=True)
        for result in query:
            db.delete(result)

        query = GaeMessageMetadata.all(keys_only=True)
        for result in query:
            db.delete(result)

        query = GaeMessageBody.all(keys_only=True)
        for result in query:
            db.delete(result)

    def subscribe(self, connection_id, client_id, topic, sub_topic=None, selector=None):
        """Add a subscription to a topic."""

        topic = SubscriptionManager.getTopicKey(topic, sub_topic)
        key_name = GaeSubscription.getKeyName(connection_id, client_id, topic)
        subscription = GaeSubscription(key_name=key_name,
            connection_id=connection_id, client_id=client_id, topic=topic)
        subscription.put()

    def unSubscribe(self, connection_id, client_id, topic, sub_topic=None):
        """Remove a subscription from a topic."""

        topic = SubscriptionManager.getTopicKey(topic, sub_topic)
        key_name = GaeSubscription.getKeyName(connection_id, client_id, topic)

        subscription = GaeSubscription.get_by_key_name(key_name)
        db.delete(subscription)

    def deleteConnection(self, connection):
        """Delete connection-subscription information."""

        query = GaeSubscription.all(keys_only=True)
        query.filter('connection_id = ', connection.id)
        db.delete(query)

    def persistMessage(self, msg):
        """Save message object."""

        # Remove connection data,
        # so that it is not pickled
        tmp_connection = getattr(msg, 'connection', None)
        if tmp_connection is not None:
            msg.connection = None

        message_body = GaeMessageBody(p_message=pickle.dumps(msg))
        message_body.put()
       
        message_data = GaeMessageMetadata(timestamp=msg.timestamp,
            time_to_live=float(msg.timeToLive), topic=self.getMessageTopicKey(msg),
            message_body=message_body)
        message_data.put()

        # Restore connection attr.
        if tmp_connection is not None:
            msg.connection = tmp_connection

    def iterConnectionSubscriptions(self, connection):
        """Iterate through all Subscriptions that belong to a specific connection."""

        query = GaeSubscription.all()
        query.filter('connection_id = ', connection.id)
        return query

    def iterSubscribers(self, topic, sub_topic=None):
        """Iterate through all connection ids subscribed to a topic."""

        topic = SubscriptionManager.getTopicKey(topic, sub_topic)

        connection_ids = {} # Keep track of unique IDs.

        query = GaeSubscription.all()
        query.filter('topic = ', topic)
        for subscription in query:
            if subscription.connection_id in connection_ids:
                continue
            
            connection_ids[subscription.connection_id] = True
            yield subscription.connection_id

    def pollMessages(self, topic, cutoff_time, current_time):
        """Retrieves all qeued messages, and discards expired messages.

        arguments:
        ===========
         * topic - string, Topic to find messages for.
         * cutoff_time - float, epoch time, only messages published
             after this time will be returned.
         * current_time - float, epoch time, used to determine if a
             message is expired.
        """

        query = GaeMessageMetadata.all()
        query.filter('topic = ', topic)
        query.filter('timestamp > ', cutoff_time)
        query.order('timestamp')
        for message_data in query:
            yield pickle.loads(message_data.message_body.p_message) 

    def deleteExpiredMessages(self, cutoff_time):
        query = GaeMessageMetadata.all(keys_only=True)
        query.filter('timestamp < ', cutoff_time)
        for result in query:
            db.delete(result)



# -*- coding: utf-8 -*-
import unittest

from StringIO import StringIO

from amfast.context import DecoderContext
import amfast.decode as decode
import amfast.buffer as buffer
import amfast.class_def as class_def

class Amf3DecoderTestCase(unittest.TestCase):
    class Spam(object):
        def __init__(self):
            self.spam = 'eggs'

    def setUp(self):
        self.class_mapper = class_def.ClassDefMapper()

    def testNone(self):
        self.assertEquals(None, decode.decode(DecoderContext('\x01', amf3=True)))

    def testFalse(self):
        self.assertEquals(False, decode.decode(DecoderContext('\x02', amf3=True)))

    def testTrue(self):
        self.assertEquals(True, decode.decode(DecoderContext('\x03', amf3=True)))

    def testInt(self):
        tests = {
            0: '\x04\x00',
            0x35: '\x04\x35',
            0x7f: '\x04\x7f',
            0x80: '\x04\x81\x00',
            0xd4: '\x04\x81\x54',
            0x3fff: '\x04\xff\x7f',
            0x4000: '\x04\x81\x80\x00',
            0x1a53f: '\x04\x86\xca\x3f',
            0x1fffff: '\x04\xff\xff\x7f',
            0x200000: '\x04\x80\xc0\x80\x00',
            -0x01: '\x04\xff\xff\xff\xff',
            -0x2a: '\x04\xff\xff\xff\xd6',
            0xfffffff: '\x04\xbf\xff\xff\xff',
            -0x10000000: '\x04\xc0\x80\x80\x00'
        }

        for integer, encoding in tests.iteritems():
            self.assertEquals(integer, decode.decode(DecoderContext(encoding, amf3=True)))

    def testEncodeFloat(self):
        tests = {
            0.1: '\x05\x3f\xb9\x99\x99\x99\x99\x99\x9a',
            0.123456789: '\x05\x3f\xbf\x9a\xdd\x37\x39\x63\x5f'
        }

        for number, encoding in tests.iteritems():
            self.assertEquals(number, decode.decode(DecoderContext(encoding, amf3=True)))

    def testEncodeLong(self):
        tests = {
            0x10000000: '\x05\x41\xb0\x00\x00\x00\x00\x00\x00',
            -0x10000001: '\x05\xc1\xb0\x00\x00\x01\x00\x00\x00',
            -0x10000000: '\x04\xc0\x80\x80\x00'
        }

        for number, encoding in tests.iteritems():
            self.assertEquals(number, decode.decode(DecoderContext(encoding, amf3=True)))

    def testUnicode(self):
        tests = {
            u'': '\x06\x01',
            u'hello': '\x06\x0bhello',
            u'ᚠᛇᚻ': '\x06\x13\xe1\x9a\xa0\xe1\x9b\x87\xe1\x9a\xbb'
        }

        for string, encoding in tests.iteritems():
            self.assertEquals(string, decode.decode(DecoderContext(encoding, amf3=True)))

    def testUnicodeRefs(self):
        test = ['hello', 'hello', 'hello', 'hello']

        encoded = '\x09\x09\x01' # array header
        encoded  += '\x06\x0bhello' # array element 1 (hello encoded)
        encoded += '\x06\x00' #array element 2 (reference to hello)
        encoded += '\x06\x00' #array element 3 (reference to hello)
        encoded += '\x06\x00' #array element 4 (reference to hello)

        results = decode.decode(DecoderContext(encoded, amf3=True))
        for result in results:
            self.assertEquals('hello', result)

    def testArray(self):
        encoded = '\x09\x09\x01' #array header
        encoded += '\x04\x00' #array element 1
        encoded += '\x04\x01' #array element 2
        encoded += '\x04\x02' #array element 3
        encoded += '\x04\x03' #array element 4

        result = decode.decode(DecoderContext(encoded, amf3=True))
        for i, val in enumerate(result):
            self.assertEquals(i, val);

    def testXml(self):
        import xml.dom.minidom

        encoded = '\x0B' # XML header
        encoded += '\x81\x2B' # String header
        encoded += '<?xml version="1.0" ?><test>\n            <test_me>tester</test_me>\n           </test>' # encoded XML

        result = decode.decode(DecoderContext(encoded, amf3=True))
        self.assertEquals(xml.dom.minidom.Document, result.__class__)

    def testEncodeXmlRef(self):
        import xml.dom.minidom
        encoded = '\x09\x05\x01' #array header
        encoded += '\x0B\x81\x2B<?xml version="1.0" ?><test>\n            <test_me>tester</test_me>\n           </test>' # array element 1 (encoded dom)
        encoded += '\x0B\x02' # array element 2 (reference to dom)

        result = decode.decode(DecoderContext(encoded, amf3=True))
        self.assertEquals(xml.dom.minidom.Document, result[0].__class__)
        self.assertEquals(result[0], result[1])

    def testDict(self):
        encoded = '\x0A\x0B\x01' # Object header
        encoded += '\x09spam' # key
        encoded += '\x06\x09eggs' #value
        encoded += '\x01' # empty string terminator

        result = decode.decode(DecoderContext(encoded, amf3=True))
        self.assertEquals('eggs', result['spam']);

    def testObjectProxy(self):
        encoded = '\x0A\x07\x3Bflex.messaging.io.ObjectProxy' # Object header 
        encoded += '\x0A\x0B\x01' # Object header
        encoded += '\x09spam' # key
        encoded += '\x06\x09eggs' #value
        encoded += '\x01' # empty string terminator

        result = decode.decode(DecoderContext(encoded, amf3=True))
        self.assertEquals('eggs', result['spam']);

    def testDictRef(self):
        encoded = '\x09\x05\x01' #array header
        encoded += '\x0A\x0B\x01\x09spam\x06\x09eggs\x01' #array element 1 (dict encoded)
        encoded += '\x0A\x02' # array element 2 (reference to dict)

        result = decode.decode(DecoderContext(encoded, amf3=True))
        self.assertEquals('eggs', result[0]['spam'])
        self.assertEquals(result[0], result[1])

    def testArrayCollection(self):
        encoded = '\x0A\x07\x43flex.messaging.io.ArrayCollection' # Object header 
        encoded += '\x09\x09\x01' #array header
        encoded += '\x04\x00' #array element 1
        encoded += '\x04\x01' #array element 2
        encoded += '\x04\x02' #array element 3
        encoded += '\x04\x03' #array element 4

        result = decode.decode(DecoderContext(encoded, amf3=True))
        for i in range(4):
            self.assertEquals(i, result[i])

        result = decode.decode(DecoderContext(StringIO(encoded), amf3=True))
        for i in range(4):
            self.assertEquals(i, result[i])

    def testArrayCollectionRef(self):
        encoded = '\x0A\x07\x43flex.messaging.io.ArrayCollection\x09\x05\x01' # array header
        encoded += '\x0A\x01\x09\x09\x01\x04\x00\x04\x01\x04\x02\x04\x03' # array element 1 (test_tuple encoded)
        encoded += '\x0A\x04' # array element 2 (reference to test_tuple)

        result = decode.decode(DecoderContext(encoded, amf3=True))
        self.assertEquals([].__class__, result.__class__)
        self.assertEquals(result[0], result[1])
        for i in range(4):
            self.assertEquals(i, result[0][i])

    def testClassRef(self):
        self.class_mapper.mapClass(class_def.ClassDef(self.Spam, 'alias.spam', ('spam',)))

        encoded = '\x09\x05\x01' #array header
        encoded += '\x0A\x13\x15alias.spam\x09spam\x06\x09eggs' # array element 1
        encoded += '\x0A\x01\x06\x07foo' # array element 2

        result = decode.decode(DecoderContext(encoded,
            class_def_mapper=self.class_mapper, amf3=True))
        self.class_mapper.unmapClass(self.Spam)

        self.assertEquals(result[0].__class__, result[1].__class__)

    def testStaticObj(self):
        self.class_mapper.mapClass(class_def.ClassDef(self.Spam, 'alias.spam', ('spam',)))

        encoded = '\x0A\x13\x15alias.spam'
        encoded += '\x09spam' # static attr definition
        encoded += '\x06\x09eggs' # static attrs

        result = decode.decode(DecoderContext(encoded,
            class_def_mapper=self.class_mapper, amf3=True))
        self.assertEquals('eggs', result.spam)

        result = decode.decode(DecoderContext(StringIO(encoded),
            class_def_mapper=self.class_mapper, amf3=True))
        self.assertEquals('eggs', result.spam)

        self.class_mapper.unmapClass(self.Spam)

    def testDynamicObj(self):
        self.class_mapper.mapClass(class_def.DynamicClassDef(self.Spam, 'alias.spam', ()))

        encoded = '\x0A\x0B\x15alias.spam'
        encoded += '\x09spam\x06\x09eggs\x01' # dynamic attrs

        result = decode.decode(DecoderContext(encoded,
            class_def_mapper=self.class_mapper, amf3=True))
        self.assertEquals('eggs', result.spam)

        result = decode.decode(DecoderContext(StringIO(encoded),
            class_def_mapper=self.class_mapper, amf3=True))
        self.assertEquals('eggs', result.spam)

        self.class_mapper.unmapClass(self.Spam)

    def testStaticDynamicObj(self):
        self.class_mapper.mapClass(class_def.DynamicClassDef(self.Spam, 'alias.spam', ('spam',)))

        test = self.Spam()
        test.ham = 'foo'

        encoded = '\x0A\x1B\x15alias.spam' # obj header
        encoded += '\x09spam' # static attr definition
        encoded += '\x06\x09eggs' # static attrs
        encoded += '\x07ham\x06\x07foo\x01' #dynamic attrs

        result = decode.decode(DecoderContext(encoded,
            class_def_mapper=self.class_mapper, amf3=True))
        self.assertEquals('eggs', result.spam)
        self.assertEquals('foo', result.ham)

        result = decode.decode(DecoderContext(StringIO(encoded),
            class_def_mapper=self.class_mapper, amf3=True))
        self.assertEquals('eggs', result.spam)
        self.assertEquals('foo', result.ham)

        self.class_mapper.unmapClass(self.Spam)

    def testExternizeable(self):
        custom_encoding = '\x01\x02\x03\x04\x05'

        class ExternClass(class_def.ExternClassDef):
            def readExternal(ext_self, obj, context):
                byte_string = context.buffer.read(5);
                self.assertEquals(custom_encoding, byte_string)
                obj.spam = 'eggs'

        self.class_mapper.mapClass(ExternClass(self.Spam, 'alias.spam', ('spam',)))

        encoded = '\x0A\x07\x15alias.spam'
        encoded += custom_encoding # raw bytes

        result = decode.decode(DecoderContext(encoded,
            class_def_mapper=self.class_mapper, amf3=True))
        self.class_mapper.unmapClass(self.Spam)

        self.assertEquals('eggs', result.spam)

    def testDate(self):
        import datetime
        encoded = '\x08\x01Bp+6!\x15\x80\x00'
        result = decode.decode(DecoderContext(encoded, amf3=True))
        self.assertEquals(2005, result.year)
        self.assertEquals(3, result.month)
        self.assertEquals(18, result.day)
        self.assertEquals(1, result.hour)
        self.assertEquals(58, result.minute)
        self.assertEquals(31, result.second)

        encoded = '\x08\x01Bo%\xe2\xb2\x80\x00\x00'
        result = decode.decode(DecoderContext(encoded, amf3=True))
        self.assertEquals(2003, result.year)
        self.assertEquals(12, result.month)
        self.assertEquals(1, result.day)

    def testUnkownByteRaisesException(self):
        self.assertRaises(decode.DecodeError, decode.decode, DecoderContext('\x0D'))

    def testBadArgRaisesException(self):
        self.assertRaises(decode.DecodeError, decode.decode, 1)

    def testReusableDecoderContext(self):
        import amfast.encoder
        encoder = amfast.encoder.Encoder(amf3=True)
        pre = {'foo' : 'bar'}
        post = encoder.encode(pre)
        ct = DecoderContext(post, amf3=True)

        result = decode.decode(ct)
        assert result == pre

        self.assertRaises(buffer.BufferError, decode.decode, ct) 

def suite():
    return unittest.TestLoader().loadTestsFromTestCase(Amf3DecoderTestCase)

if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())

